# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Mosaics area bounded by bounding boxes over detected object
"""

from typing import Any, Dict, List

import cv2
import numpy as np

from peekingduck.pipeline.nodes.node import AbstractNode


class Node(AbstractNode):  # pylint: disable=too-few-public-methods
    """Mosaics areas bounded by bounding boxes on image.

    The :mod:`draw.mosaic_bbox` node helps to anonymize detected objects by
    pixelating the areas bounded by bounding boxes in an image.

    Inputs:
        |img_data|

        |bboxes_data|

    Outputs:
        |img_data|

    Configs:
        mosaic_level (:obj:`int`): **default = 7**. |br|
            Defines the resolution of a mosaic filter (width |times| height).
            The number corresponds to the number of rows and columns used to
            create a mosaic. For example, the default setting
            (``mosaic_level = 7``) creates a :math:`7 \\times 7` mosaic filter.
            Increasing the number increases the intensity of pixelization over
            an area.
    """

    def __init__(self, config: Dict[str, Any] = None, **kwargs: Any) -> None:
        super().__init__(config, node_path=__name__, **kwargs)

        self.mosaic_level = self.config["mosaic_level"]

    def run(self, inputs: Dict[str, Any]) -> Dict[str, Any]:
        mosaic_img = self._mosaic_bbox(inputs["img"], inputs["bboxes"])
        outputs = {"img": mosaic_img}
        return outputs

    def _mosaic_bbox(self, image: np.ndarray, bboxes: List[np.ndarray]) -> np.ndarray:
        """Mosaics areas bounded by bounding boxes on ``image``.

        Args:
            image (np.ndarray): Image in numpy array.
            bboxes (List[np.ndarray]): numpy array of detected bboxes

        Returns:
            (np.ndarray): Image with mosaicked bounding box regions.
        """
        height, width = image.shape[:2]
        # Prevent calculating mosaic on a mosaicked area
        original_image = image.copy()

        for bbox in bboxes:
            # bbox can contain negative values sometimes, ensures the ROI
            # selection is within bounds and without wrapping
            rows = slice(int(max(0, bbox[1]) * height), int(min(1, bbox[3]) * height))
            cols = slice(int(max(0, bbox[0]) * width), int(min(1, bbox[2]) * width))

            image[rows, cols] = self._mosaic(original_image[rows, cols])

        return image

    def _mosaic(self, image: np.ndarray) -> np.ndarray:
        """Mosaics a given input image.

        Args:
            image (np.ndarray): Image in numpy array.

        Returns:
            (np.ndarray): Mosaicked image in numpy array.
        """
        height, width = image.shape[:2]
        image = cv2.resize(
            image,
            (self.mosaic_level, self.mosaic_level),
            interpolation=cv2.INTER_LANCZOS4,
        )
        image = cv2.resize(image, (width, height), interpolation=cv2.INTER_NEAREST)

        return image



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""PoseNet model with model types: mobilenet50, mobilenet75, mobilenet100 and resnet"""

import logging
from typing import Dict, Any, Tuple
import numpy as np

from peekingduck.weights_utils import checker, downloader, finder
from peekingduck.pipeline.nodes.model.posenetv1.posenet_files.predictor import Predictor


class PoseNetModel:  # pylint: disable=too-few-public-methods
    """PoseNet model with model types: mobilenet50, mobilenet75, mobilenet100 and resnet"""

    def __init__(self, config: Dict[str, Any]) -> None:
        super().__init__()

        self.logger = logging.getLogger(__name__)

        # check threshold values
        if not 0 <= config["score_threshold"] <= 1:
            raise ValueError("score_threshold must be in [0, 1]")
        if config["model_type"] not in [50, 75, 100, "resnet"]:
            raise ValueError("model_type must be one of [50, 75, 100, resnet]")

        weights_dir, model_dir = finder.find_paths(
            config["root"], config["weights"], config["weights_parent_dir"]
        )

        # check for posenet weights, if none then download into weights folder
        if not checker.has_weights(weights_dir, model_dir):
            self.logger.info("---no weights detected. proceeding to download...---")
            downloader.download_weights(weights_dir, config["weights"]["blob_file"])
            self.logger.info(f"---weights downloaded to {weights_dir}.---")

        self.predictor = Predictor(config, model_dir)

    def predict(
        self, frame: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """ Predict poses from input frame

        Args:
            frame (np.array): image in numpy array

        Returns:
            bboxes, keypoints, keypoint_scores, keypoint_masks, keypoint_conns
            (Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]): \
            tuple containing list of bboxes and pose related info i.e coordinates,
            scores, connections
        """
        assert isinstance(frame, np.ndarray)

        return self.predictor.predict(frame)



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
EfficientDet model with model types: D0-D4
"""

import logging
from pathlib import Path
from typing import Any, Dict, List, Tuple
import json
import numpy as np

from peekingduck.pipeline.nodes.model.efficientdet_d04.efficientdet_files.detector import (
    Detector,
)
from peekingduck.weights_utils import checker, downloader, finder


class EfficientDetModel:
    """EfficientDet model with model types: D0-D4"""

    def __init__(self, config: Dict[str, Any]) -> None:
        super().__init__()

        self.logger = logging.getLogger(__name__)

        # check threshold values
        if not 0 <= config["score_threshold"] <= 1:
            raise ValueError("score_threshold must be in [0, 1]")
        if not 0 <= config["model_type"] <= 4:
            raise ValueError("model_type must be an integer in [0, 4]")

        # check for efficientdet weights, if none then download into weights folder
        weights_dir, model_dir = finder.find_paths(
            config["root"], config["weights"], config["weights_parent_dir"]
        )
        if not checker.has_weights(weights_dir, model_dir):
            self.logger.info("---no weights detected. proceeding to download...---")
            downloader.download_weights(weights_dir, config["weights"]["blob_file"])
            self.logger.info(f"---weights downloaded to {weights_dir}.---")

        classes_path = model_dir / config["weights"]["classes_file"]
        self.class_names = {
            val["id"] - 1: val["name"]
            for val in json.loads(Path(classes_path).read_text()).values()
        }
        self.detector = Detector(config, model_dir, self.class_names)
        self.detect_ids = config["detect_ids"]

    @property
    def detect_ids(self) -> List[int]:
        """The list of selected object category IDs."""
        return self._detect_ids

    @detect_ids.setter
    def detect_ids(self, ids: List[int]) -> None:
        if not isinstance(ids, list):
            raise TypeError("detect_ids has to be a list")
        if not ids:
            self.logger.info("Detecting all EfficientDet classes")
        self._detect_ids = ids

    def predict(
        self, image: np.ndarray
    ) -> Tuple[List[np.ndarray], List[str], List[float]]:
        """predict the bbox from frame

        Args:
            image (np.ndarray): Input image frame.

        Returns:
            object_bboxes(List[Numpy ndarray]): list of bboxes detected
            object_labels(List[str]): list of index labels of the
                object detected for the corresponding bbox
            object_scores(List[float]): list of confidence scores of the
                object detected for the corresponding bbox
        """
        if not isinstance(image, np.ndarray):
            raise TypeError("image must be a np.ndarray")

        # returns object_bboxes, object_labels, object_scores
        return self.detector.predict_object_bbox_from_image(image, self.detect_ids)



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Functions to download model weights.
"""

import os
import zipfile
from pathlib import Path

import requests
from tqdm import tqdm

BASE_URL = "https://storage.googleapis.com/peekingduck/models"


def download_weights(weights_dir: Path, blob_file: str) -> None:
    """Downloads weights for specified ``blob_file``.

    Args:
        weights_dir (:obj:`Path`): Path to where all weights are stored.
        blob_file (:obj:`str`): Name of file to be downloaded.
    """
    zip_path = weights_dir / "temp.zip"

    download_file_from_blob(blob_file, zip_path)

    # search for downloaded .zip file and extract, then delete
    with zipfile.ZipFile(zip_path, "r") as temp:
        for file in tqdm(iterable=temp.namelist(), total=len(temp.namelist())):
            temp.extract(member=file, path=weights_dir)

    os.remove(zip_path)


def download_file_from_blob(file_name: str, destination: Path) -> None:
    """Downloads publicly shared files from Google Cloud Platform.

    Args:
        file_name (:obj:`str`): Name of file to be downloaded.
        destination (:obj:`Path`): Destination directory of download.
    """
    url = f"{BASE_URL}/{file_name}"
    session = requests.Session()

    response = session.get(url, stream=True)
    save_response_content(response, destination)


def save_response_content(response: requests.Response, destination: Path) -> None:
    """Saves download content in chunks.

    Chunk size set to large integer as weights are usually pretty large.

    Args:
        response (:obj:`requests.Response`): HTML response.
        destination (:obj:`Path`): Destination directory of download.
    """
    chunk_size = 32768

    with open(destination, "wb") as temp:
        for chunk in tqdm(response.iter_content(chunk_size)):
            if chunk:  # filter out keep-alive new chunks
                temp.write(chunk)



# Modifications copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Original copyright (c) 2019 ZhongdaoWang
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Functions for bounding box matching.

Modifications include:
- Pure python replacement of cython_bbox
- Removed checking for List[np.ndarray] types in iou_distance()
- Set return_cost=False and use list comprehension in linear_assignment()
- Removed only_position argument in fuse_motion as only False value is used.
"""

from typing import List, Tuple

import lap
import numpy as np
from scipy.spatial.distance import cdist

from peekingduck.pipeline.nodes.model.jdev1.jde_files.kalman_filter import (
    KalmanFilter,
    chi2inv95,
)
from peekingduck.pipeline.nodes.model.jdev1.jde_files.track import STrack


def bbox_ious(bboxes_1: np.ndarray, bboxes_2: np.ndarray) -> np.ndarray:
    """Calculates the Intersection-over-Union (IoU) between bounding boxes.
    Bounding boxes have the format (x1, y1, x2, y2), where (x1, y1) is the
    top-left corner and (x2, y2) is the bottom-right corner. The algorithm is
    adapted from simple-faster-rcnn-pytorch with modifications such as adding
    1 to area calculations to match the equations used in cython_bbox.

    Args:
        bboxes_1 (np.ndarray): An array with shape (N, 4) where N is the number
            of bounding boxes.
        bboxes_2 (np.ndarray): An array similar to `bboxes_2` with shape (K, 4)
            where K is the number of bounding boxes.

    Returns:
        (np.ndarray): An array with shape (N, K). The element at index (n, k)
        contains the IoU between the n-th bounding box in `bboxes_1` and the
        k-th bounding box in `bboxes_2`.

    Reference:
        simple-faster-rcnn-pytorch
        https://github.com/chenyuntc/simple-faster-rcnn-pytorch

        cython_bbox:
        https://github.com/samson-wang/cython_bbox
    """
    # top left
    intersect_tl = np.maximum(bboxes_1[:, np.newaxis, :2], bboxes_2[:, :2])
    # bottom right
    intersect_br = np.minimum(bboxes_1[:, np.newaxis, 2:], bboxes_2[:, 2:]) + 1

    intersect_area = np.prod(intersect_br - intersect_tl, axis=2) * (
        intersect_tl < intersect_br
    ).all(axis=2)
    area_1 = np.prod(bboxes_1[:, 2:] - bboxes_1[:, :2] + 1, axis=1)
    area_2 = np.prod(bboxes_2[:, 2:] - bboxes_2[:, :2] + 1, axis=1)
    iou_values = intersect_area / (area_1[:, np.newaxis] + area_2 - intersect_area)

    return iou_values


def embedding_distance(
    tracks: List[STrack], detections: List[STrack], metric: str = "euclidean"
) -> np.ndarray:
    """Computes cost based on features between `tracks` and `detections`.

    Args:
        tracks (List[STrack]): List of STracks.
        detections (List[STrack]): List of STracks that are model predictions.
        metric (str): The metric to be used with
            `scipy.spatial.distance.cdist()`. Defaults to "euclidean".

    Returns:
        np.ndarray: Cost matrix of distance.
    """
    cost_matrix = np.zeros((len(tracks), len(detections)), dtype=np.float)
    if cost_matrix.size == 0:
        return cost_matrix
    det_features = np.asarray([track.curr_feat for track in detections], dtype=np.float)
    track_features = np.asarray([track.smooth_feat for track in tracks], dtype=np.float)
    # Normalized features
    cost_matrix = np.maximum(0.0, cdist(track_features, det_features, metric))

    return cost_matrix


def fuse_motion(  # pylint: disable=too-many-arguments
    kalman_filter: KalmanFilter,
    cost_matrix: np.ndarray,
    tracks: List[STrack],
    detections: List[STrack],
    coeff: float = 0.98,
) -> np.ndarray:
    """Computes the cost matrix using the pair-wise motion affinity matrix and
    appearance affinity matrix.

    Args:
        kalman_filter (KalmanFilter): Kalman filter for state estimation.
        cost_matrix (np.ndarray): Cost matrix filled with values from the
            appearance affinity matrix.
        tracks (List[STrack]): List of STracks.
        detections (List[STrack]): List of STracks that are model predictions.
        coeff (float): Weighting parameter used in computing the final cost
            matrix, corresponds to `lambda` in the arxiv article.

    Returns:
        (np.ndarray): Cost matrix used by Hungarian algorithm to solve the
        linear assignment problem.
    """
    if cost_matrix.size == 0:
        return cost_matrix
    gating_threshold = chi2inv95[4]
    measurements = np.asarray([det.xyah for det in detections])
    for row, track in enumerate(tracks):
        gating_distance = kalman_filter.gating_distance(
            track.mean, track.covariance, measurements
        )
        cost_matrix[row, gating_distance > gating_threshold] = np.inf
        cost_matrix[row] = coeff * cost_matrix[row] + (1 - coeff) * gating_distance
    return cost_matrix


def iou_distance(tracks_1: List[STrack], tracks_2: List[STrack]) -> np.ndarray:
    """Computes cost based on Intersection-over-Union (IoU).

    Args:
        tracks_1 (List[STrack]): List of STracks.
        tracks_2 (List[STrack]): List of STracks.

    Returns:
        (np.ndarray): Cost matrix of distance between IoU of bounding boxes.
    """
    xyxys_1 = [track.xyxy for track in tracks_1]
    xyxys_2 = [track.xyxy for track in tracks_2]
    cost_matrix = 1 - ious(xyxys_1, xyxys_2)

    return cost_matrix


def ious(xyxys_1: List[np.ndarray], xyxys_2: List[np.ndarray]) -> np.ndarray:
    """Computes a matrix Intersection-over-Union (IoU) values between 2 list
    of bounding boxes with (x1, y1, x2, y2) format where (x1, y1) is the top
    left and (x2, y2) is the bottom right.

    Args:
        xyxys_1 (List[np.ndarray]): List of STracks.
        xyxys_2 (List[np.ndarray]): List of STracks.

    Returns:
        np.ndarray: Matrix of IoU values.
    """
    iou_values = np.zeros((len(xyxys_1), len(xyxys_2)), dtype=np.float)
    if iou_values.size == 0:
        return iou_values

    return bbox_ious(
        np.ascontiguousarray(xyxys_1, dtype=np.float),
        np.ascontiguousarray(xyxys_2, dtype=np.float),
    )


def linear_assignment(
    cost_matrix: np.ndarray, threshold: float
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Uses Hungarian Algorithm to associate detections to tracks.

    Args:
        cost_matrix (np.ndarray): Cost matrix which is a weighted sum of the
            pair-wise motion affinity matrix and appearance affinity matrix.
        threshold (float): An upper limit for a cost of a single assignment.

    Returns:
        (Tuple[np.ndarray, np.ndarray, np.ndarray]): Returned tuple
            contains arrays of matched and unmatched tracks.
    """
    if cost_matrix.size == 0:
        return (
            np.empty((0, 2), dtype=int),
            np.arange(cost_matrix.shape[0], dtype=int),
            np.arange(cost_matrix.shape[1], dtype=int),
        )
    x_assignment, y_assignment = lap.lapjv(
        cost_matrix, extend_cost=True, cost_limit=threshold, return_cost=False
    )
    matches = np.asarray(
        [[row, col] for row, col in enumerate(x_assignment) if col >= 0]
    )
    unmatched_1 = np.where(x_assignment < 0)[0]
    unmatched_2 = np.where(y_assignment < 0)[0]
    return matches, unmatched_1, unmatched_2



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Writes the output image/video to file.
"""

import datetime
from pathlib import Path
from typing import Any, Dict, Optional

import cv2
import numpy as np

from peekingduck.pipeline.nodes.node import AbstractNode

# role of this node is to be able to take in multiple frames, stitch them
# together and output them.
# to do: need to have 'live' kind of data when there is no filename
# to do: it will be good to have the accepted file format as a configuration
# to do: somewhere so that input and output can use this config for media related issues


class Node(AbstractNode):
    """Outputs the processed image or video to a file. A timestamp is appended to the
    end of the file name.

    Inputs:
        |img_data|

        |filename_data|

        |saved_video_fps_data|

        |pipeline_end_data|

    Outputs:
        |none_output_data|

    Configs:
        output_dir (:obj:`str`): **default = "PeekingDuck/data/output"**. |br|
            Output directory for files to be written locally.
    """

    def __init__(self, config: Dict[str, Any] = None, **kwargs: Any) -> None:
        super().__init__(config, node_path=__name__, **kwargs)

        self.output_dir = Path(self.output_dir)  # type: ignore
        self._file_name: Optional[str] = None
        self._file_path_with_timestamp: Optional[str] = None
        self._image_type: Optional[str] = None
        self.writer = None
        self._prepare_directory(self.output_dir)
        self._fourcc = cv2.VideoWriter_fourcc(*"mp4v")
        self.logger.info(f"Output directory used is: {self.output_dir}")

    def run(self, inputs: Dict[str, Any]) -> Dict[str, Any]:
        """Writes media information to filepath."""
        # reset and terminate when there are no more data
        if inputs["pipeline_end"]:
            if self.writer:  # images automatically releases writer
                self.writer.release()
            return {}
        if not self._file_name:
            self._prepare_writer(
                inputs["filename"], inputs["img"], inputs["saved_video_fps"]
            )
        if inputs["filename"] != self._file_name:
            self._prepare_writer(
                inputs["filename"], inputs["img"], inputs["saved_video_fps"]
            )
        self._write(inputs["img"])

        return {}

    def _write(self, img: np.ndarray) -> None:
        if self._image_type == "image":
            cv2.imwrite(self._file_path_with_timestamp, img)
        else:
            self.writer.write(img)

    def _prepare_writer(
        self, filename: str, img: np.ndarray, saved_video_fps: int
    ) -> None:
        self._file_path_with_timestamp = self._append_datetime_filename(filename)

        if filename.split(".")[-1] in ["jpg", "jpeg", "png"]:
            self._image_type = "image"
        else:
            self._image_type = "video"
            resolution = img.shape[1], img.shape[0]
            self.writer = cv2.VideoWriter(
                self._file_path_with_timestamp,
                self._fourcc,
                saved_video_fps,
                resolution,
            )

    @staticmethod
    def _prepare_directory(output_dir: Path) -> None:
        output_dir.mkdir(parents=True, exist_ok=True)

    def _append_datetime_filename(self, filename: str) -> str:
        self._file_name = filename
        current_time = datetime.datetime.now()
        # output as 'YYYYMMDD_hhmmss'
        time_str = current_time.strftime("%y%m%d_%H%M%S")

        # append timestamp to filename before extension Format: filename_timestamp.extension
        filename_with_timestamp = f"_{time_str}.".join(filename.split(".")[-2:])
        file_path_with_timestamp = self.output_dir / filename_with_timestamp

        return str(file_path_with_timestamp)



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Adjusts the brightness of an image.
"""


from typing import Any, Dict

import cv2
import numpy as np

from peekingduck.pipeline.nodes.node import AbstractNode


class Node(AbstractNode):
    """Adjusts the brightness of an image, by adding a bias/`beta parameter
    <https://docs.opencv.org/4.x/d3/dc1/tutorial_basic_
    linear_transform.html>`_.

    Inputs:
        |img_data|

    Outputs:
        |img_data|

    Configs:
        beta (:obj:`int`): **[-100, 100], default = 0**. |br|
            Increasing the value of beta increases image brightness, and vice
            versa.
    """

    def __init__(self, config: Dict[str, Any] = None, **kwargs: Any) -> None:
        super().__init__(config, node_path=__name__, **kwargs)

        if not -100 <= self.beta <= 100:
            raise ValueError("beta for brightness must be between [-100, 100]")

    def run(self, inputs: Dict[str, Any]) -> Dict[str, Any]:
        """Adjusts the brightness of an image frame.

        Args:
            inputs (Dict): Inputs dictionary with the key `img`.

        Returns:
            (Dict): Outputs dictionary with the key `img`.
        """
        orig_shape = inputs["img"].shape
        img_vector = np.reshape(inputs["img"], (1, -1))
        cv2.add(img_vector, self.beta, img_vector)
        img = np.reshape(img_vector, orig_shape)

        return {"img": img}



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""🎯 Performs multiple object tracking for detected bboxes."""

from typing import Any, Dict

from peekingduck.pipeline.nodes.dabble.trackingv1.detection_tracker import (
    DetectionTracker,
)
from peekingduck.pipeline.nodes.node import AbstractNode


class Node(AbstractNode):
    """Uses bounding boxes detected by an object detector model to track
    multiple objects. :mod:`dabble.tracking` is a useful alternative to
    :mod:`model.fairmot` and :mod:`model.jde` as it can track bounding boxes
    detected by the upstream object detector and is not limited to only
    ``"person"`` detections.

    Currently, two types of tracking algorithms can be selected: MOSSE and IOU.
    Information on the algorithms' performance can be found
    :ref:`here <object-tracking-benchmarks>`.

    Inputs:
        |img_data|

        |bboxes_data|

    Outputs:
        |obj_attrs_data|
        :mod:`dabble.tracking` produces the ``ids`` attribute which contains
        the tracking IDs of the detections.


    Configs:
        tracking_type (:obj:`str`): **{"iou", "mosse"}, default="iou"**. |br|
            Type of tracking algorithm to be used. For more information about
            the trackers, please view the :doc:`Object Counting (Over Time) use
            case </use_cases/object_counting_over_time>`.
        iou_threshold (:obj:`float`): **[0, 1], default=0.1**. |br|
            Minimum IoU value to be used with the matching logic.
        max_lost (:obj:`int`): **[0, sys.maxsize), default=10**. |br|
            Maximum number of frames to keep "lost" tracks after which they
            will be removed. Only used when ``tracking_type = iou``.
    """

    def __init__(self, config: Dict[str, Any] = None, **kwargs: Any) -> None:
        super().__init__(config, node_path=__name__, **kwargs)
        self.tracker = DetectionTracker(self.config)

    def run(self, inputs: Dict[str, Any]) -> Dict[str, Any]:
        """Tracks detection bounding boxes.

        Args:
            inputs (Dict[str, Any]): Dictionary with keys "img", "bboxes", and
                "bbox_scores.

        Returns:
            outputs (Dict[str, Any]): Tracking IDs of bounding boxes.
            "obj_attrs" key is used for compatibility with draw nodes.
        """
        # Potentially use frame_rate here too since IOUTracker has a
        # max_time_lost
        metadata = inputs.get("mot_metadata", {"reset_model": False})
        reset_model = metadata["reset_model"]
        if reset_model:
            self._reset_model()

        track_ids = self.tracker.track_detections(inputs)

        return {"obj_attrs": {"ids": track_ids}}

    def _reset_model(self) -> None:
        """Creates a new instance of DetectionTracker."""
        self.logger.info(f"Creating new {self.config['tracking_type']} tracker...")
        self.tracker = DetectionTracker(self.config)



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Constants used for PoseNet
"""

import numpy as np

KEYPOINT_NAMES = (
    "nose",
    "leftEye",
    "rightEye",
    "leftEar",
    "rightEar",
    "leftShoulder",
    "rightShoulder",
    "leftElbow",
    "rightElbow",
    "leftWrist",
    "rightWrist",
    "leftHip",
    "rightHip",
    "leftKnee",
    "rightKnee",
    "leftAnkle",
    "rightAnkle",
)

_KEYPOINT_IDS = {pn: pid for pid, pn in enumerate(KEYPOINT_NAMES)}

_POSE_CHAIN = [
    ("nose", "leftEye"),
    ("leftEye", "leftEar"),
    ("nose", "rightEye"),
    ("rightEye", "rightEar"),
    ("nose", "leftShoulder"),
    ("leftShoulder", "leftElbow"),
    ("leftElbow", "leftWrist"),
    ("leftShoulder", "leftHip"),
    ("leftHip", "leftKnee"),
    ("leftKnee", "leftAnkle"),
    ("nose", "rightShoulder"),
    ("rightShoulder", "rightElbow"),
    ("rightElbow", "rightWrist"),
    ("rightShoulder", "rightHip"),
    ("rightHip", "rightKnee"),
    ("rightKnee", "rightAnkle"),
]

POSE_CONNECTIONS = np.array(
    [(_KEYPOINT_IDS[parent], _KEYPOINT_IDS[child]) for parent, child in _POSE_CHAIN]
)

# fmt: off
SKELETON = [
    [16, 14], [14, 12], [17, 15], [15, 13], [12, 13], [6, 12], [7, 13], [6, 7],
    [6, 8], [7, 9], [8, 10], [9, 11], [2, 3], [1, 2], [1, 3], [2, 4], [3, 5],
    [4, 6], [5, 7],
]
# fmt: on

LOCAL_MAXIMUM_RADIUS = 1

SCALE_FACTOR = 1.0

KEYPOINTS_NUM = 17

MIN_PART_SCORE = 0.1

IMAGE_NET_MEAN = [-123.15, -115.90, -103.06]

SWAP_AXES = (1, 0)



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
QuickFind algorithm used to connect nodes
"""

from typing import List


class QuickFind:
    """
    The general form of the quick-find algorithm that is used to connect nodes.
    How it works: https://www.cs.princeton.edu/~rs/AlgsDS07/01UnionFind.pdf
    Note that although there are faster algorithms such as union-find, we are using
    quick-find here because we need the returning array to have the final group numbers,
    and not have to continue to chase the roots of the nodes subsequently.
    """

    def __init__(self, arr_size: int) -> None:
        self.group_alloc = []
        self.arr_size = arr_size
        for i in range(arr_size):
            self.group_alloc.append(i)

    def get_group_alloc(self) -> List[int]:
        """Getter function for self.group_alloc

        Returns:
            (list): a list containing group numbers allocated to each node.
        """
        return self.group_alloc

    def union(self, node_1_idx: int, node_2_idx: int) -> None:
        """Unites two nodes and other nodes that have been connected to them.

        Args:
            node_1_idx (int): index of the first node
            node_2_idx (int): index of the second node
        """

        node_1_group = self.group_alloc[node_1_idx]
        node_2_group = self.group_alloc[node_2_idx]
        for i in range(self.arr_size):
            if self.group_alloc[i] == node_1_group:
                self.group_alloc[i] = node_2_group

    def connected(self, node_1_idx: int, node_2_idx: int) -> bool:
        """Checks if two nodes are connected.

        Args:
            node_1_idx (int): index of the first node
            node_2_idx (int): index of the second node

        Returns:
            (bool): True if 2 nodes are connected
        """
        return self.group_alloc[node_1_idx] == self.group_alloc[node_2_idx]



# Copyright 2018 Ross Wightman
#
# Modifications copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Support functions to decode single pose
"""


from typing import Tuple
import numpy as np

from peekingduck.pipeline.nodes.model.posenetv1.posenet_files.constants import (
    POSE_CONNECTIONS,
)


def decode_pose(
    root_score: float,
    root_id: int,
    root_image_coord: np.ndarray,
    scores: np.ndarray,
    offsets: np.ndarray,
    output_stride: int,
    displacements_fwd: np.ndarray,
    displacements_bwd: np.ndarray,
    keypoint_scores: np.ndarray,
    keypoint_coords: np.ndarray,
) -> None:
    # pylint: disable=too-many-arguments
    """Decode pose's keypoints scores and coordinates from keypoints score,
    coordinates and displacements
    Args:
        root_score (float): a keypoint with highest score is selected as root
        root_id (int): root keypoint's index
        root_image_coord (np.array): relative coordinate of root keypoint
        scores (np.array): HxWxNP heatmap scores of NP body parts
        offsets (np.array): HxWxNPx2 short range offset vector of NP body parts
        output_stride (int): output stride to convert output indices to image coordinates
        displacements_fwd (np.array): HxWxNEx2 forward displacements of NE body
                connections
        displacements_bwd (np.array): HxWxNEx2 backward displacements of NE body
                connections
        keypoints_scores (np.array): 17x1 buffer to store keypoint scores
        keypoint_coords (np.array): 17x2 buffer to store keypoint coordinates
    Returns:
        pose_count (int): number of poses detected
    """
    num_edges = len(POSE_CONNECTIONS)

    keypoint_scores[root_id] = root_score
    keypoint_coords[root_id] = root_image_coord

    for edge in reversed(range(num_edges)):
        target_keypoint_id, source_keypoint_id = POSE_CONNECTIONS[edge]
        _calculate_instance_keypoints(
            edge,
            target_keypoint_id,
            source_keypoint_id,
            keypoint_scores,
            keypoint_coords,
            scores,
            offsets,
            output_stride,
            displacements_bwd,
        )

    for edge in range(num_edges):
        source_keypoint_id, target_keypoint_id = POSE_CONNECTIONS[edge]
        _calculate_instance_keypoints(
            edge,
            target_keypoint_id,
            source_keypoint_id,
            keypoint_scores,
            keypoint_coords,
            scores,
            offsets,
            output_stride,
            displacements_fwd,
        )


def _calculate_instance_keypoints(
    edge: int,
    target_keypoint_id: int,
    source_keypoint_id: int,
    instance_keypoint_scores: np.ndarray,
    instance_keypoint_coords: np.ndarray,
    scores: np.ndarray,
    offsets: np.ndarray,
    output_stride: int,
    displacements: np.ndarray,
) -> None:
    # pylint: disable=too-many-arguments
    """Obtain instance keypoints scores and coordinates"""
    if (
        instance_keypoint_scores[source_keypoint_id] > 0.0
        and instance_keypoint_scores[target_keypoint_id] == 0.0
    ):
        source_keypoint = instance_keypoint_coords[source_keypoint_id]

        score, coords = _traverse_to_target_keypoint(
            edge,
            source_keypoint,
            target_keypoint_id,
            scores,
            offsets,
            output_stride,
            displacements,
        )

        instance_keypoint_scores[target_keypoint_id] = score
        instance_keypoint_coords[target_keypoint_id] = coords


def _clip_to_indices(
    keypoints: np.ndarray, output_stride: int, width: int, height: int
) -> np.ndarray:
    """Clip keypoint coordinate to indices within dimension (width, height)"""
    keypoints = keypoints / output_stride
    keypoint_indices = np.zeros((2,), dtype=np.int32)

    keypoint_indices[0] = max(min(round(keypoints[0]), width - 1), 0)
    keypoint_indices[1] = max(min(round(keypoints[1]), height - 1), 0)

    return keypoint_indices


def _traverse_to_target_keypoint(
    edge_id: int,
    source_keypoint: np.ndarray,
    target_keypoint_id: int,
    scores: np.ndarray,
    offsets: np.ndarray,
    output_stride: int,
    displacements: np.ndarray,
) -> Tuple[float, float]:
    # pylint: disable=too-many-arguments
    """Traverse to target keypoint to obtain keypoint score and coordinates"""
    height = scores.shape[0] - 1
    width = scores.shape[1] - 1

    source_keypoint_indices = _clip_to_indices(
        source_keypoint, output_stride, width, height
    )

    displaced_point = (
        source_keypoint
        + displacements[source_keypoint_indices[1], source_keypoint_indices[0], edge_id]
    )

    displaced_point_indices = _clip_to_indices(
        displaced_point, output_stride, width, height
    )

    score = scores[
        displaced_point_indices[1], displaced_point_indices[0], target_keypoint_id
    ]

    image_coord = (
        displaced_point_indices * output_stride
        + offsets[
            displaced_point_indices[1], displaced_point_indices[0], target_keypoint_id
        ]
    )

    return score, image_coord



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Displays selected information from preceding nodes in a legend box.
"""

from typing import Any, Dict, List

from peekingduck.pipeline.nodes.draw.utils.legend import Legend
from peekingduck.pipeline.nodes.node import AbstractNode


class Node(AbstractNode):
    """Draws a translucent legend box on a corner of the image, containing
    selected information produced by preceding nodes in the format
    ``<data type>: <value>``. Supports in-built PeekingDuck data types defined
    in :doc:`Glossary </glossary>` as well as custom data types produced by
    custom nodes.

    This example screenshot shows :term:`fps` from :mod:`dabble.fps`,
    :term:`count` from :mod:`dabble.bbox_count` and :term:`cum_avg` from
    :mod:`dabble.statistics` displayed within the legend box.

    .. image:: /assets/api/legend.png

    |br|

    With the exception of the :term:`zone_count` data type from
    :mod:`dabble.zone_count`, all other selected in-built PeekingDuck data
    types or custom data types must be of types :obj:`int`, :obj:`float`, or
    :obj:`str`. Note that values of float type such as :term:`fps` and
    :term:`cum_avg` are displayed in 2 decimal places.

    Inputs:
        |all_input_data|

    Outputs:
        |img_data|

    Configs:
        position (:obj:`str`): **{"top", "bottom"}, default = "bottom"**. |br|
            Position to draw legend box. "top" draws it at the top-left
            position while "bottom" draws it at bottom-left.
        show (:obj:`List[str]`): **default = []**. |br|
            Include in this list the desired data type(s) to be drawn within
            the legend box, such as ``["fps", "count", "cum_avg"]`` in the
            example screenshot. Custom data types produced by custom nodes are
            also supported. If no data types are included, an error will be
            produced.

    .. versionchanged:: 1.2.0
        Merged previous ``all_legend_items`` and ``include`` configs into a
        single ``show`` config for greater clarity. Added support for drawing
        custom data types produced by custom nodes, to improve the flexibility
        of this node.
    """

    def __init__(self, config: Dict[str, Any] = None, **kwargs: Any) -> None:
        super().__init__(config, node_path=__name__, **kwargs)
        if not self.show:
            raise KeyError(
                "To display information in the legend box, at least one data type must be "
                "selected in the 'show' config."
            )

    def run(self, inputs: Dict[str, Any]) -> Dict[str, Any]:
        """Draws legend box with information from nodes.

        Args:
            inputs (dict): Dictionary with all available keys.

        Returns:
            outputs (dict): Dictionary with keys "none".
        """
        _check_data_type(inputs, self.show)
        Legend().draw(inputs, self.show, self.position)
        # cv2 weighted does not update the referenced image. Need to return and replace.
        return {"img": inputs["img"]}


def _check_data_type(inputs: Dict[str, Any], show: List[str]) -> None:
    """Checks if the data types provided in show were produced from preceding nodes."""

    for item in show:
        if item not in inputs:
            raise KeyError(
                f"'{item}' was selected for drawing, but is not a valid data type from preceding "
                f"nodes."
            )



# Modifications copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Original copyright 2021 Megvii, Base Detection
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Backbone for YOLOPAFPN.

Modifications include:
- Removed unused Darknet class
- Removed unused DWConv, ResLayer import
- Removed depthwise and act arguments
- Refactor and formatting
"""

from typing import Dict, Tuple

import torch
import torch.nn as nn

from peekingduck.pipeline.nodes.model.yoloxv1.yolox_files.network_blocks import (
    BaseConv,
    CSPLayer,
    Focus,
    SPPBottleneck,
)


class CSPDarknet(nn.Module):
    """Modified CSPNet with SiLU activation.

    Args:
        dep_mul (float): Depth multiplier, used to determine the number of
            Bottlenecks in `CSPLayer`.
        wid_mul (float): Width multiplier, used to determine the number of
            `in_channels` and `out_channels` in `BaseConv`.
        out_features (Tuple[str, str, str]): Selects the desired outputs in
            `forward()`. `YOLOPAFPN` creates `CSPDarknet` with
            `out_features = ("dark3", "dark4", "dark5")`.
    """

    # pylint: disable=arguments-differ
    def __init__(
        self,
        dep_mul: float,
        wid_mul: float,
        out_features: Tuple[str, str, str],
    ) -> None:
        super().__init__()
        self.out_features = out_features

        channels = int(wid_mul * 64)  # 64
        depth = max(round(dep_mul * 3), 1)  # 3

        self.stem = Focus(3, channels, 3)
        self.dark2 = nn.Sequential(*self.make_group_layer(channels, depth))
        self.dark3 = nn.Sequential(*self.make_group_layer(channels * 2, depth * 3))
        self.dark4 = nn.Sequential(*self.make_group_layer(channels * 4, depth * 3))
        self.dark5 = nn.Sequential(
            BaseConv(channels * 8, channels * 16, 3, 2),
            SPPBottleneck(channels * 16, channels * 16),
            CSPLayer(channels * 16, channels * 16, depth, False),
        )

    @staticmethod
    def make_group_layer(in_channels: int, depth: int) -> Tuple[BaseConv, CSPLayer]:
        """Starts with BaseConv layer, followed by a CSPLayer.

        Args:
            in_channels (int): Number of channels in the input image.
            depth (int): Number of Bottlenecks.

        Returns:
            (Tuple[BaseConv, CSPLayer]): A group layer consisting of a BaseConv
            CSPLayer.
        """
        return (
            BaseConv(in_channels, in_channels * 2, 3, 2),
            CSPLayer(in_channels * 2, in_channels * 2, depth),
        )

    def forward(self, inputs: torch.Tensor) -> Dict[str, torch.Tensor]:
        """Defines the computation performed at every call.

        Args:
            inputs (torch.Tensor): Input from the previous layer.

        Returns:
            (Dict[str, torch.Tensor]): A dictionary of tensors with keys
            corresponding to `self.out_features`.
        """
        outputs = {}
        inputs = self.stem(inputs)
        outputs["stem"] = inputs
        inputs = self.dark2(inputs)
        outputs["dark2"] = inputs
        inputs = self.dark3(inputs)
        outputs["dark3"] = inputs
        inputs = self.dark4(inputs)
        outputs["dark4"] = inputs
        inputs = self.dark5(inputs)
        outputs["dark5"] = inputs
        return {k: v for k, v in outputs.items() if k in self.out_features}



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Pipeline class that stores nodes and manages the data information used during
inference.
"""

import textwrap
from typing import Any, Dict, List

from peekingduck.pipeline.nodes.node import AbstractNode


class Pipeline:  # pylint: disable=too-few-public-methods
    """Pipeline class that stores nodes and manages flow of data used during
    inference.

    Args:
        nodes (:obj:`List[AbstractNode]`): List of initialized nodes for the
            pipeline to run through.
    """

    def __init__(self, nodes: List[AbstractNode]) -> None:
        self.nodes = nodes
        self._check_pipe(nodes)
        self.data = {}  # type: ignore
        self.terminate = False

    def get_pipeline_results(self) -> Dict[str, Any]:
        """Gets all results data from nodes in pipeline.

        Returns:
            (:obj:`Dict[str, Any]`): Dictionary of all results from nodes in
            the pipeline.
        """
        return self.data

    @staticmethod
    def _check_pipe(nodes: List[AbstractNode]) -> None:
        # 1. Check the initial node is a source node
        # 2. Check every subsequent node utilizes something that will exist
        # if reached end, it is all valid
        data_pool = []

        if nodes[0].inputs[0] == "none":
            data_pool.extend(nodes[0].outputs)

        for node in nodes[1:]:
            if all(item in data_pool for item in node.inputs) or "all" in node.inputs:
                data_pool.extend(node.outputs)
            else:
                msg = textwrap.dedent(
                    f"""\
                    Nodes in this pipeline do not form a proper channel:
                    {node.name} requires these inputs: {node.inputs}
                    Data pool only has these outputs from previous nodes: {data_pool}
                    Note that nodes run sequentially, in the order specified in the config file.
                    """
                )
                raise ValueError(msg)



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Superimposes a heat map over an image."""

from typing import Any, Dict

import cv2
import numpy as np

from peekingduck.pipeline.nodes.node import AbstractNode


class Node(AbstractNode):  # pylint: disable=too-few-public-methods
    """Superimposes a heat map over an image.

    The :mod:`draw.heat_map` node helps to identify areas that are more
    crowded. Areas that are more crowded are highlighted in red while areas
    that are less crowded are highlighted in blue.

    Inputs:
        |img_data|

        |density_map_data|
        This is produced by nodes such as :mod:`model.csrnet`.

    Outputs:
        |img_data|

    Configs:
        None.
    """

    def __init__(self, config: Dict[str, Any] = None, **kwargs: Any) -> None:
        super().__init__(config, node_path=__name__, **kwargs)

    def run(self, inputs: Dict[str, Any]) -> Dict[str, Any]:
        heat_map_img = self._add_heat_map(inputs["density_map"], inputs["img"])
        outputs = {"img": heat_map_img}
        return outputs

    def _add_heat_map(self, density_map: np.ndarray, image: np.ndarray) -> np.ndarray:
        """Superimposes a heat map over an ``image``.

        Args:
            density_map (np.ndarray): predicted density map.
            image (np.ndarray): image in numpy array.

        Returns:
            image (np.ndarray): image with a heat map superimposed over it.
        """
        if np.count_nonzero(density_map) != 0:
            density_map = self._norm_min_max(density_map)
            heat_map = cv2.applyColorMap(density_map, cv2.COLORMAP_JET)
            image = cv2.addWeighted(image, 0.5, heat_map, 0.5, 0)

        return image

    @staticmethod
    def _norm_min_max(src: np.ndarray) -> np.ndarray:
        target = None
        norm_results = cv2.normalize(
            # source array
            src,
            # destination array
            target,
            # lower boundary value
            alpha=0,
            # upper boundary value
            beta=255,
            # normalization type
            norm_type=cv2.NORM_MINMAX,
            # data type
            dtype=cv2.CV_8U,
        )
        return norm_results



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Helper classes to load configurations and nodes.
"""

import ast
import collections.abc
import importlib
import logging
import sys
from pathlib import Path
from typing import Any, Dict, Iterator, List, Optional, Tuple, Union

import yaml

from peekingduck.configloader import ConfigLoader
from peekingduck.pipeline.nodes.node import AbstractNode
from peekingduck.pipeline.pipeline import Pipeline
from peekingduck.utils.create_node_helper import obj_det_change_class_name_to_id

PEEKINGDUCK_NODE_TYPES = ["input", "augment", "model", "draw", "dabble", "output"]


class DeclarativeLoader:  # pylint: disable=too-few-public-methods
    """A helper class to create
    :py:class:`Pipeline <peekingduck.pipeline.pipeline.Pipeline>`.

    The declarative loader class creates the specified nodes according to any
    modifications provided in the configs and returns the pipeline needed for
    inference.

    Args:
        pipeline_path (:obj:`pathlib.Path`): Path to a YAML file that
            declares the node sequence to be used in the pipeline.
        config_updates_cli (:obj:`str`): Stringified nested dictionaries of
            configuration changes passed as part of CLI command. Used to modify
            the node configurations directly from the CLI.
        custom_nodes_parent_subdir (:obj:`str`): Relative path to parent
            folder which contains custom nodes that users have created to be
            used with PeekingDuck. For more information on using custom nodes,
            please refer to
            `Getting Started <getting_started/03_custom_nodes.html>`_.
    """

    def __init__(
        self,
        pipeline_path: Path,
        config_updates_cli: str,
        custom_nodes_parent_subdir: str,
    ) -> None:
        self.logger = logging.getLogger(__name__)

        self.pkd_base_dir = Path(__file__).resolve().parent
        self.config_loader = ConfigLoader(self.pkd_base_dir)

        self.node_list = self._load_node_list(pipeline_path)
        self.config_updates_cli = ast.literal_eval(config_updates_cli)

        custom_nodes_name = self._get_custom_name_from_node_list()
        if custom_nodes_name is not None:
            custom_nodes_dir = (
                Path.cwd() / custom_nodes_parent_subdir / custom_nodes_name
            )
            self.custom_config_loader = ConfigLoader(custom_nodes_dir)
            sys.path.append(custom_nodes_parent_subdir)

            self.custom_nodes_dir = custom_nodes_dir

    def _load_node_list(self, pipeline_path: Path) -> "NodeList":
        """Loads a list of nodes from pipeline_path.yml"""

        # dotw 2022-03-17: Temporary helper methods
        def deprecation_warning(name: str, config: Union[str, Dict[str, Any]]) -> None:
            self.logger.warning(f"`{name}` deprecated, replaced by `input.visual`")
            self.logger.warning(f"convert `{name}` to `input.visual`:{config}")

        with open(pipeline_path) as node_yml:
            data = yaml.safe_load(node_yml)
        if not isinstance(data, dict) or "nodes" not in data:
            raise ValueError(
                f"{pipeline_path} has an invalid structure. "
                "Missing top-level 'nodes' key."
            )

        nodes = data["nodes"]
        if nodes is None:
            raise ValueError(f"{pipeline_path} does not contain any nodes!")

        # dotw 2022-03-16: Temporary code to convert existing `input.live` and
        #                  `input.recorded` into new `input.visual`
        #                  To be removed in subsequent versions
        upgraded_nodes = []
        for node in nodes:
            if isinstance(node, str):
                if node in ["input.live", "input.recorded"]:
                    deprecation_warning(node, "input.visual")
                    if node == "input.live":
                        node = {"input.visual": {"source": 0}}
                    else:
                        self.logger.error("input.recorded with no parameters error!")
                        node = "input.visual"
            else:
                if "input.live" in node:
                    node_config = node.pop("input.live")
                    if "input_source" in node_config:
                        node_config["source"] = node_config.pop("input_source")
                    node["input.visual"] = node_config
                    deprecation_warning("input.live", node_config)
                if "input.recorded" in node:
                    node_config = node.pop("input.recorded")
                    if "input_dir" in node_config:
                        node_config["source"] = node_config.pop("input_dir")
                    node["input.visual"] = node_config
                    deprecation_warning("input.recorded", node_config)
            upgraded_nodes.append(node)

        self.logger.info("Successfully loaded pipeline file.")
        return NodeList(upgraded_nodes)

    def _get_custom_name_from_node_list(self) -> Any:
        custom_name = None

        for node_str, _ in self.node_list:
            node_type = node_str.split(".")[0]

            if node_type not in PEEKINGDUCK_NODE_TYPES:
                custom_name = node_type
                break

        return custom_name

    def _instantiate_nodes(self) -> List[AbstractNode]:
        """Given a list of imported nodes, instantiate nodes"""
        instantiated_nodes = []

        for node_str, config_updates_yml in self.node_list:
            node_str_split = node_str.split(".")

            self.logger.info(f"Initializing {node_str} node...")

            if len(node_str_split) == 3:
                # convert windows/linux filepath to a module path
                path_to_node = f"{self.custom_nodes_dir.name}."
                node_name = ".".join(node_str_split[-2:])

                instantiated_node = self._init_node(
                    path_to_node,
                    node_name,
                    self.custom_config_loader,
                    config_updates_yml,
                )
            else:
                path_to_node = "peekingduck.pipeline.nodes."

                instantiated_node = self._init_node(
                    path_to_node, node_str, self.config_loader, config_updates_yml
                )

            instantiated_nodes.append(instantiated_node)

        return instantiated_nodes

    def _init_node(
        self,
        path_to_node: str,
        node_name: str,
        config_loader: ConfigLoader,
        config_updates_yml: Optional[Dict[str, Any]],
    ) -> AbstractNode:
        """Imports node to filepath and initializes node with config."""
        node = importlib.import_module(path_to_node + node_name)
        config = config_loader.get(node_name)

        # First, override default configs with values from pipeline_config.yml
        if config_updates_yml is not None:
            config = self._edit_config(config, config_updates_yml, node_name)

        # Second, override configs again with values from cli
        if self.config_updates_cli is not None:
            if node_name in self.config_updates_cli.keys():
                config = self._edit_config(
                    config, self.config_updates_cli[node_name], node_name
                )

        return node.Node(config)

    def _edit_config(
        self, dict_orig: Dict[str, Any], dict_update: Dict[str, Any], node_name: str
    ) -> Dict[str, Any]:
        """Update value of a nested dictionary of varying depth using recursion"""
        for key, value in dict_update.items():
            if isinstance(value, collections.abc.Mapping):
                dict_orig[key] = self._edit_config(
                    dict_orig.get(key, {}), value, node_name  # type: ignore
                )
            else:
                if key not in dict_orig:
                    self.logger.warning(
                        f"Config for node {node_name} does not have the key: {key}"
                    )
                else:
                    if key == "detect_ids":
                        key, value = obj_det_change_class_name_to_id(
                            node_name, key, value
                        )

                    dict_orig[key] = value
                    self.logger.info(
                        f"Config for node {node_name} is updated to: '{key}': {value}"
                    )
        return dict_orig

    def get_pipeline(self) -> Pipeline:
        """Returns a compiled
        :py:class:`Pipeline <peekingduck.pipeline.pipeline.Pipeline>` for
        PeekingDuck :py:class:`Runner <peekingduck.runner.Runner>` to execute.
        """
        instantiated_nodes = self._instantiate_nodes()

        try:
            return Pipeline(instantiated_nodes)
        except ValueError as error:
            self.logger.error(str(error))
            sys.exit(1)


class NodeList:
    """Iterator class to return node string and node configs (if any) from the
    nodes declared in the run config file.
    """

    def __init__(self, nodes: List[Union[Dict[str, Any], str]]) -> None:
        self.nodes = nodes
        self.length = len(nodes)

    def __iter__(self) -> Iterator[Tuple[str, Optional[Dict[str, Any]]]]:
        self.current = -1
        return self

    def __next__(self) -> Tuple[str, Optional[Dict[str, Any]]]:
        self.current += 1
        if self.current >= self.length:
            raise StopIteration
        node_item = self.nodes[self.current]

        if isinstance(node_item, dict):
            node_str = next(iter(node_item))
            config_updates = node_item[node_str]
        else:
            node_str = node_item
            config_updates = None

        return node_str, config_updates



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Draws utilities regarding zoning functions
"""

from typing import Any, List, Tuple

import cv2
import numpy as np

from peekingduck.pipeline.nodes.draw.utils.constants import (
    PRIMARY_PALETTE,
    PRIMARY_PALETTE_LENGTH,
    VERY_THICK,
)


def _draw_zone_area(
    frame: np.ndarray, points: List[Tuple[int]], zone_index: int
) -> None:
    num_points = len(points)
    for i in range(num_points):
        cv2.line(
            frame,
            points[i],
            points[(i + 1) % num_points],
            PRIMARY_PALETTE[(zone_index + 1) % PRIMARY_PALETTE_LENGTH],
            VERY_THICK,
        )


def draw_zones(frame: np.ndarray, zones: List[Any]) -> None:
    """draw the boundaries of the zones used in zoning analytics

    Args:
        frame (np.array): image of current frame
        zones (Zone): zones used in the zoning analytics. possible
        classes are Area and Divider.
    """
    for i, zone_pts in enumerate(zones):
        _draw_zone_area(frame, zone_pts, i)



# Modifications copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Original copyright (c) 2019 ZhongdaoWang
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Kalman Filter for updating Track motion states.

Modifications include:
- Removed unused distance metric in gating_distance()
- Removed predict() as only multi_predict() is used by JDE
- Removed only_position argument in gating_distance() as only False value is
    used
"""

from typing import Tuple

import numpy as np
import scipy.linalg

# Table for the 0.95 quantile of the chi-square distribution with N degrees of
# freedom (contains values for N=1, ..., 9). Taken from MATLAB/Octave's chi2inv
# function and used as Mahalanobis gating threshold.
chi2inv95 = {
    1: 3.8415,
    2: 5.9915,
    3: 7.8147,
    4: 9.4877,
    5: 11.070,
    6: 12.592,
    7: 14.067,
    8: 15.507,
    9: 16.919,
}


class KalmanFilter:
    """A simple Kalman filter for tracking bounding boxes in image space.

    The 8-dimensional state space

        x, y, a, h, vx, vy, va, vh

    contains the bounding box center position (x, y), aspect ratio a, height h,
    and their respective velocities.

    Object motion follows a constant velocity model. The bounding box location
    (x, y, a, h) is taken as direct observation of the state space (linear
    observation model).
    """

    num_dims = 4
    dt = 1.0

    def __init__(self) -> None:
        # Create Kalman filter model matrices.
        self._motion_mat = np.eye(2 * self.num_dims, 2 * self.num_dims)
        for i in range(self.num_dims):
            self._motion_mat[i, self.num_dims + i] = self.dt
        self._update_mat = np.eye(self.num_dims, 2 * self.num_dims)

        # Motion and observation uncertainty are chosen relative to the current
        # state estimate. These weights control the amount of uncertainty in
        # the model. This is a bit hacky.
        self._std_weight_position = 1.0 / 20
        self._std_weight_velocity = 1.0 / 160

    def gating_distance(
        self,
        mean: np.ndarray,
        covariance: np.ndarray,
        measurements: np.ndarray,
    ) -> np.ndarray:
        """Computes gating distance between state distribution and
        measurements using Mahalanobis distance.

        A suitable distance threshold can be obtained from `chi2inv95`. If
        `only_position` is False, the chi-square distribution has 4 degrees of
        freedom, otherwise 2.

        Args:
            mean (np.ndarray): Mean vector over the state distribution (8
                dimensional).
            covariance (np.ndarray): Covariance of the state distribution (8x8
                dimensional).
            measurements (np.ndarray): An Nx4 dimensional matrix of N
                measurements, each in format (x, y, a, h) where (x, y) is the
                bounding box center position, a the aspect ratio, and h the
                height.

        Returns:
            (np.ndarray): An array of length N, where the i-th element contains
            the squared Mahalanobis distance between (mean, covariance) and
            `measurements[i]`.
        """
        mean, covariance = self.project(mean, covariance)

        distances = measurements - mean
        cholesky_factor = np.linalg.cholesky(covariance)
        maha_distance = scipy.linalg.solve_triangular(
            cholesky_factor,
            distances.T,
            lower=True,
            overwrite_b=True,
            check_finite=False,
        )
        squared_maha = np.sum(maha_distance * maha_distance, axis=0)
        return squared_maha

    def initiate(self, measurement: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """Creates track from unassociated measurement.

        Args:
            measurement (np.ndarray): Bounding box coordinates (x, y, a, h)
                with center position (x, y), aspect ratio a, and height h.

        Returns:
            (Tuple[np.ndarray, np.ndarray]): The mean vector (8 dimensional)
            and covariance matrix (8x8 dimensional) of the new track.
            Unobserved velocities are initialized to 0 mean.
        """
        mean_pos = measurement
        mean_vel = np.zeros_like(mean_pos)
        mean = np.r_[mean_pos, mean_vel]

        std = [
            2 * self._std_weight_position * measurement[3],
            2 * self._std_weight_position * measurement[3],
            1e-2,
            2 * self._std_weight_position * measurement[3],
            10 * self._std_weight_velocity * measurement[3],
            10 * self._std_weight_velocity * measurement[3],
            1e-5,
            10 * self._std_weight_velocity * measurement[3],
        ]
        covariance = np.diag(np.square(std))
        return mean, covariance

    def multi_predict(
        self, mean: np.ndarray, covariance: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray]:
        """Runs Kalman filter prediction step (Vectorized version).

        Args:
            mean (np.ndarray): The Nx8 dimensional mean matrix of the object
                states at the previous time step.
            covariance (np.ndarray): The Nx8x8 dimensional covariance matrices
                of the object states at the previous time step.

        Returns:
            (Tuple[np.ndarray, np.ndarray]): The mean vector and covariance
            matrix of the predicted state. Unobserved velocities are
            initialized to 0 mean.
        """
        std_pos = [
            self._std_weight_position * mean[:, 3],
            self._std_weight_position * mean[:, 3],
            1e-2 * np.ones_like(mean[:, 3]),
            self._std_weight_position * mean[:, 3],
        ]
        std_vel = [
            self._std_weight_velocity * mean[:, 3],
            self._std_weight_velocity * mean[:, 3],
            1e-5 * np.ones_like(mean[:, 3]),
            self._std_weight_velocity * mean[:, 3],
        ]
        sqr = np.square(np.r_[std_pos, std_vel]).T

        motion_cov = []
        for i in range(len(mean)):
            motion_cov.append(np.diag(sqr[i]))
        motion_cov = np.asarray(motion_cov)

        mean = np.dot(mean, self._motion_mat.T)
        left = np.dot(self._motion_mat, covariance).transpose((1, 0, 2))
        covariance = np.dot(left, self._motion_mat.T) + motion_cov

        return mean, covariance

    def project(
        self, mean: np.ndarray, covariance: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray]:
        """Projects state distribution to measurement space.

        Args:
            mean (np.ndarray): The state's mean vector (8 dimensional array).
            covariance (np.ndarray): The state's covariance matrix (8x8 dimensional).

        Returns:
            (Tuple[np.ndarray, np.ndarray]): The projected mean and covariance
            matrix of the given state estimate.
        """
        std = [
            self._std_weight_position * mean[3],
            self._std_weight_position * mean[3],
            1e-1,
            self._std_weight_position * mean[3],
        ]
        innovation_cov = np.diag(np.square(std))

        mean = np.dot(self._update_mat, mean)
        covariance = np.linalg.multi_dot(
            (self._update_mat, covariance, self._update_mat.T)
        )
        return mean, covariance + innovation_cov

    def update(
        self, mean: np.ndarray, covariance: np.ndarray, measurement: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray]:
        """Runs Kalman filter correction step.

        Args:
            mean (np.ndarray): The predicted state's mean vector (8
                dimensional).
            covariance (np.ndarray): The state's covariance matrix (8x8
                dimensional).
            measurement (np.ndarray): The 4 dimensional measurement vector
                (x, y, a, h), where (x, y) is the center position, a the aspect
                ratio, and h the height of the bounding box.

        Returns:
            (Tuple[np.ndarray, np.ndarray]): The measurement-corrected state
            distribution.
        """
        projected_mean, projected_cov = self.project(mean, covariance)

        chol_factor, lower = scipy.linalg.cho_factor(
            projected_cov, lower=True, check_finite=False
        )
        kalman_gain = scipy.linalg.cho_solve(
            (chol_factor, lower),
            np.dot(covariance, self._update_mat.T).T,
            check_finite=False,
        ).T
        innovation = measurement - projected_mean

        new_mean = mean + np.dot(innovation, kalman_gain.T)
        new_covariance = covariance - np.linalg.multi_dot(
            (kalman_gain, projected_cov, kalman_gain.T)
        )
        return new_mean, new_covariance



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
License Plate detection model with model types: yolov4 and yolov4tiny
"""

import logging
from typing import Any, Dict, Tuple

import numpy as np

from peekingduck.pipeline.nodes.model.yolov4_license_plate.licenseplate_files.detector import (
    Detector,
)
from peekingduck.weights_utils import checker, downloader, finder


class Yolov4:  # pylint: disable=too-few-public-methods
    """Yolo model with model types: v4 and v4tiny"""

    def __init__(self, config: Dict[str, Any]) -> None:
        super().__init__()

        self.logger = logging.getLogger(__name__)

        # check threshold values
        if not 0 <= config["iou_threshold"] <= 1:
            raise ValueError("iou_threshold must be in [0, 1]")

        if not 0 <= config["score_threshold"] <= 1:
            raise ValueError("score_threshold must be in [0, 1]")

        weights_dir, model_dir = finder.find_paths(
            config["root"], config["weights"], config["weights_parent_dir"]
        )

        # check for yolo weights, if none then download into weights folder
        if not checker.has_weights(weights_dir, model_dir):
            self.logger.info("---no weights detected. proceeding to download...---")
            downloader.download_weights(weights_dir, config["weights"]["blob_file"])
            self.logger.info(f"---weights downloaded to {weights_dir}.---")

        self.detector = Detector(config, model_dir)

    def predict(self, frame: np.ndarray) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """
        Predicts the bboxes from image frame

        Returns:
                object_bboxes (Numpy Array): list of bboxes detected
                object_labels (Numpy Array): list of string labels of the
                    object detected for the corresponding bbox
                object_scores (Numpy Array): list of confidence scores of the
                    object detected for the corresponding bbox
        """
        assert isinstance(frame, np.ndarray)

        return self.detector.predict_object_bbox_from_image(frame)



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Loads configurations for individual nodes.
"""

from pathlib import Path
from typing import Any, Dict

import logging
import yaml


class ConfigLoader:  # pylint: disable=too-few-public-methods
    """A helper class to create pipeline.

    The config loader class is used to allow for instantiation of Node classes
    directly instead of reading configurations from the run config yaml.

    Args:
        base_dir (:obj:`pathlib.Path`): Base directory of ``peekingduck``
    """

    def __init__(self, base_dir: Path) -> None:
        self._base_dir = base_dir
        self.logger = logging.getLogger(__name__)

    def _get_config_path(self, node: str) -> Path:
        """Based on the node, return the corresponding node config path"""
        configs_folder = self._base_dir / "configs"
        node_type, node_name = node.split(".")
        file_path = configs_folder / node_type / f"{node_name}.yml"

        return file_path

    def get(self, node_name: str) -> Dict[str, Any]:
        """Gets node configuration for specified node.

        Args:
            node_name (:obj:`str`): Name of node.

        Returns:
            node_config (:obj:`Dict[str, Any]`): A dictionary of node
            configurations for the specified node.
        """
        file_path = self._get_config_path(node_name)

        with open(file_path) as file:
            node_config = yaml.safe_load(file)

        # some models require the knowledge of where the root is for loading
        node_config["root"] = self._base_dir
        return node_config



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Object detection class using yolo single label model
to find license plate object bboxes
"""

import logging
from pathlib import Path
from typing import Any, Dict, List, Tuple

import cv2 as cv
import numpy as np
import tensorflow as tf
from tensorflow.python.saved_model import tag_constants


class Detector:
    """Object detection class using yolo model to find object bboxes"""

    def __init__(self, config: Dict[str, Any], model_dir: Path) -> None:
        self.logger = logging.getLogger(__name__)

        self.config = config
        self.model_dir = model_dir
        self.class_labels = self._get_class_labels()
        self.yolo = self._create_yolo_model()

    def _get_class_labels(self) -> List[str]:
        classes_path = self.model_dir / self.config["weights"]["classes_file"]
        with open(classes_path, "rt") as file:
            class_labels = file.read().rstrip("\n").split("\n")

        return class_labels

    def _create_yolo_model(self) -> cv.dnn_Net:
        """
        Creates yolo model for license plate detection
        """
        self.model_type = self.config["model_type"]
        model_path = (
            self.model_dir
            / self.config["weights"]["saved_model_subdir"][self.model_type]
        )
        model = tf.saved_model.load(str(model_path), tags=[tag_constants.SERVING])

        self.logger.info(
            "Yolo model loaded with following configs: \n\t"
            f"Model type: {self.config['model_type']}, \n\t"
            f"Input resolution: {self.config['size']}, \n\t"
            f"IOU threshold: {self.config['iou_threshold']}, \n\t"
            f"Score threshold: {self.config['score_threshold']}"
        )

        return model

    @staticmethod
    def bbox_scaling(bboxes: List[list], scale_factor: float) -> List[list]:
        """
        To scale the width and height of bboxes from v4tiny
        After the conversion of the model in .cfg and .weight file format, from
        Alexey's Darknet repo, to tf model, bboxes are bigger.
        So downscaling is required for a better fit
        """
        for idx, box in enumerate(bboxes):
            x_1, y_1, x_2, y_2 = tuple(box)
            center_x = (x_1 + x_2) / 2
            center_y = (y_1 + y_2) / 2
            scaled_x_1 = center_x - ((x_2 - x_1) / 2 * scale_factor)
            scaled_x_2 = center_x + ((x_2 - x_1) / 2 * scale_factor)
            scaled_y_1 = center_y - ((y_2 - y_1) / 2 * scale_factor)
            scaled_y_2 = center_y + ((y_2 - y_1) / 2 * scale_factor)
            bboxes[idx] = [scaled_x_1, scaled_y_1, scaled_x_2, scaled_y_2]

        return bboxes

    def predict_object_bbox_from_image(
        self, image: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """
        Detect all license plate objects' bounding box from one image

        args:
                image: (Numpy Array) input image

        return:
                boxes: (Numpy Array) an array of bounding box with
                    definition like (x1, y1, x2, y2), in a
                    coordinate system with origin point in
                    the top-left corner
                labels: (Numpy Array) an array of class labels
                scores: (Numpy Array) an array of confidence scores
        """
        # Use TF2 .pb saved model format for inference
        image_data = cv.cvtColor(image, cv.COLOR_BGR2RGB)
        image_data = cv.resize(image_data, (self.config["size"], self.config["size"]))
        image_data = image_data / 255.0

        image_data = np.asarray([image_data]).astype(np.float32)
        infer = self.yolo.signatures["serving_default"]
        pred_bbox = infer(tf.constant(image_data))
        for _, value in pred_bbox.items():
            pred_conf = value[:, :, 4:]
            boxes = value[:, :, 0:4]

        # performs nms using model's predictions
        bboxes, scores, classes, nums = tf.image.combined_non_max_suppression(
            tf.reshape(boxes, (tf.shape(boxes)[0], -1, 1, 4)),
            tf.reshape(
                pred_conf, (tf.shape(pred_conf)[0], -1, tf.shape(pred_conf)[-1])
            ),
            self.config["max_output_size_per_class"],
            self.config["max_total_size"],
            self.config["iou_threshold"],
            self.config["score_threshold"],
        )
        classes = classes.numpy()[0]
        classes = classes[: nums[0]]
        bboxes = bboxes.numpy()[0]
        bboxes = bboxes[: nums[0]]
        scores = scores.numpy()[0]
        scores = scores[: nums[0]]

        bboxes[:, [0, 1]] = bboxes[:, [1, 0]]  # swapping x and y axes
        bboxes[:, [2, 3]] = bboxes[:, [3, 2]]

        # scaling of bboxes if v4tiny model is used
        if self.model_type == "v4tiny":
            bboxes = self.bbox_scaling(bboxes, 0.75)

        # update the labels names of the object detected
        labels = np.asarray([self.class_labels[int(i)] for i in classes])

        return bboxes, labels, scores



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Abstract Node class for all nodes.
"""

import collections
import logging
from abc import ABCMeta, abstractmethod
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

from peekingduck.configloader import ConfigLoader
from peekingduck.utils.create_node_helper import obj_det_change_class_name_to_id


class AbstractNode(metaclass=ABCMeta):
    """Abstract Node class for inheritance by nodes.

    Defines default attributes and methods of a node.

    Args:
        config (:obj:`Dict[str, Any]` | :obj:`None`): Node configuration.
        node_path (:obj:`str`): Period-separated (``.``) relative path to the
            node from the ``peekingduck`` directory. **Default: ""**.
        pkd_base_dir (:obj:`pathlib.Path` | :obj:`None`): Path to
            ``peekingduck`` directory.
    """

    def __init__(
        self,
        config: Dict[str, Any] = None,
        node_path: str = "",
        pkd_base_dir: Optional[Path] = None,
        **kwargs: Any,
    ) -> None:
        self._name = node_path
        self.logger = logging.getLogger(self._name)

        if not pkd_base_dir:
            pkd_base_dir = Path(__file__).resolve().parents[2]

        self.node_name = ".".join(node_path.split(".")[-2:])

        # This is only initialized when the `optional_inputs` key is found in
        # the nodes' config file
        self.optional_inputs: List[str]

        # NOTE: config and kwargs_config are similar but are from different
        # inputs config is when users input a dictionary to update the node
        # kwargs_config is when users input parameters to update the node
        self.config_loader = ConfigLoader(pkd_base_dir)
        self.load_node_config(config, kwargs)  # type: ignore

        # For object detection nodes, convert class names to class ids, if any
        if self.node_name in ["model.yolo", "model.efficientdet", "model.yolox"]:
            current_ids = self.config["detect_ids"]
            _, updated_ids = obj_det_change_class_name_to_id(
                self.node_name, "detect_ids", current_ids
            )
            self.config["detect_ids"] = updated_ids

    @classmethod
    def __subclasshook__(cls: Any, subclass: Any) -> bool:
        return hasattr(subclass, "run") and callable(subclass.run)

    @abstractmethod
    def run(self, inputs: Dict[str, Any]) -> Dict[str, Any]:
        """abstract method needed for running node"""
        raise NotImplementedError("This method needs to be implemented")

    # pylint: disable=R0201, W0107
    def release_resources(self) -> None:
        """To gracefully release any acquired system resources, e.g. webcam
        NOTE: To be overridden by subclass if required"""
        pass

    @property
    def inputs(self) -> List[str]:
        """Input requirements."""
        return self.input

    @property
    def outputs(self) -> List[str]:
        """Node outputs."""
        return self.output

    @property
    def name(self) -> str:
        """Node name."""
        return self._name

    def load_node_config(
        self, config: Dict[str, Any], kwargs_config: Dict[str, Any]
    ) -> None:
        """Loads node configuration.

        NOTE: ``config`` and ``kwargs_config`` are similar but come from
        different inputs. ``config`` is when users input a dictionary to update
        the node and ``kwargs_config`` is when users input parameters to update
        the node.

        Args:
            config (:obj:`Dict[str, Any]`): Loads configuration from a
                dictionary input.
            kwargs_config (:obj:`Dict[str, Any]`): Loads configuration from
                ``kwargs``.
        """
        # if full set of configuration is not included in config
        # load configuration and update node with **kwargs where possible
        # else load from kwargs only
        self.config = config
        if not self.config:
            loaded_config = self.config_loader.get(self.node_name)
            updated_config = self._edit_config(loaded_config, kwargs_config)
            self.config = updated_config

        # sets class attributes
        for key in self.config:
            setattr(self, key, self.config[key])

    def _edit_config(
        self,
        dict_orig: Dict[str, Any],
        dict_update: Union[Dict[str, Any], collections.abc.Mapping],
    ) -> Dict[str, Any]:
        """Update value of a nested dictionary of varying depth using
        recursion
        """
        if dict_update:
            for key, value in dict_update.items():
                if isinstance(value, collections.abc.Mapping):
                    dict_orig[key] = self._edit_config(dict_orig.get(key, {}), value)
                elif key not in dict_orig:
                    self.logger.warning(
                        f"Config for node {self.node_name} does not have the key: {key}"
                    )
                else:
                    dict_orig[key] = value
                    self.logger.info(
                        f"Config for node {self.node_name} is updated to: '{key}': {value}"
                    )
        return dict_orig



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Predictor class to handle detection of poses for posenet
"""

import logging
from pathlib import Path
from typing import Any, Callable, Dict, List, Tuple

import numpy as np
import tensorflow as tf

from peekingduck.pipeline.nodes.model.posenetv1.posenet_files.constants import (
    KEYPOINTS_NUM,
    MIN_PART_SCORE,
    SCALE_FACTOR,
    SKELETON,
)
from peekingduck.pipeline.nodes.model.posenetv1.posenet_files.detector import (
    detect_keypoints,
    get_keypoints_relative_coords,
)
from peekingduck.pipeline.nodes.model.posenetv1.posenet_files.preprocessing import (
    rescale_image,
)
from peekingduck.utils.graph_functions import load_graph

OUTPUT_STRIDE = 16


class Predictor:  # pylint: disable=too-many-instance-attributes
    """Predictor class to handle detection of poses for posenet"""

    def __init__(self, config: Dict[str, Any], model_dir: Path) -> None:
        self.logger = logging.getLogger(__name__)

        self.config = config
        self.model_dir = model_dir
        self.model_type = self.config["model_type"]

        self.posenet_model = self._create_posenet_model()

    def _create_posenet_model(self) -> tf.keras.Model:
        self.resolution = self.get_resolution_as_tuple(self.config["resolution"])
        self.max_pose_detection = self.config["max_pose_detection"]
        self.score_threshold = self.config["score_threshold"]

        model_path = (
            self.model_dir / self.config["weights"]["model_file"][self.model_type]
        )
        model_func = self._load_posenet_graph(model_path)

        self.logger.info(
            "PoseNet model loaded with following configs: \n\t"
            f"Model type: {self.model_type}, \n\t"
            f"Input resolution: {self.resolution}, \n\t"
            f"Max pose detection: {self.max_pose_detection}, \n\t"
            f"Score threshold: {self.score_threshold}"
        )
        return model_func

    def _load_posenet_graph(self, model_path: Path) -> Callable:
        if self.model_type == "resnet":
            model_id = "resnet"
        else:
            model_id = "mobilenet"
        model_nodes = self.config["MODEL_NODES"][model_id]
        if model_path.is_file():
            return load_graph(
                str(model_path),
                inputs=model_nodes["inputs"],
                outputs=model_nodes["outputs"],
            )
        raise ValueError(
            "PoseNet graph file does not exist. Please check that "
            "%s exists" % model_path
        )

    @staticmethod
    def get_resolution_as_tuple(resolution: dict) -> Tuple[int, int]:
        """Convert resolution from dict to tuple format

        Args:
            resolution (dict): height and width in dict format

        Returns:
            resolution (Tuple(int)): height and width in tuple format
        """
        res1, res2 = resolution["height"], resolution["width"]

        return (int(res1), int(res2))

    def predict(
        self, frame: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        # pylint: disable=too-many-locals
        """PoseNet prediction function

        Args:
            frame (np.array): image in numpy array

        Returns:
            bboxes (np.ndarray): array of bboxes
            keypoints (np.ndarray): array of keypoint coordinates
            keypoints_scores (np.ndarray): array of keypoint scores
            keypoints_conns (np.ndarray): array of keypoint connections
        """
        (
            full_keypoint_rel_coords,
            full_keypoint_scores,
            full_masks,
        ) = self._predict_all_poses(self.posenet_model, frame, self.model_type)

        bboxes = []
        keypoints = []
        keypoint_scores = []
        keypoint_conns = []

        for pose_coords, pose_scores, pose_masks in zip(
            full_keypoint_rel_coords, full_keypoint_scores, full_masks
        ):
            bbox = self._get_bbox_of_one_pose(pose_coords, pose_masks)
            pose_coords = self._get_valid_full_keypoints_coords(pose_coords, pose_masks)
            pose_connections = self._get_connections_of_one_pose(
                pose_coords, pose_masks
            )
            bboxes.append(bbox)
            keypoints.append(pose_coords)
            keypoint_scores.append(pose_scores)
            keypoint_conns.append(pose_connections)

        return (
            np.array(bboxes),
            np.array(keypoints),
            np.array(keypoint_scores),
            np.array(keypoint_conns),
        )

    @staticmethod
    def _get_valid_full_keypoints_coords(
        coords: np.ndarray, masks: np.ndarray
    ) -> np.ndarray:
        """Apply masks to keep only valid keypoints' relative coordinates

        Args:
            coords (np.array): Nx2 array of keypoints' relative coordinates
            masks (np.array): masks for valid (> min confidence score) keypoints

        Returns:
            full_joints (np.array): Nx2 array of keypoints where undetected
                keypoints are assigned a (-1) value.
        """
        full_joints = coords.copy()
        full_joints = np.clip(full_joints, 0, 1)
        full_joints[~masks] = -1
        return full_joints

    @staticmethod
    def _get_connections_of_one_pose(
        coords: np.ndarray, masks: np.ndarray
    ) -> np.ndarray:
        """Get connections between adjacent keypoint pairs if both keypoints are
        detected
        """
        connections = []
        for start_joint, end_joint in SKELETON:
            if masks[start_joint - 1] and masks[end_joint - 1]:
                connections.append((coords[start_joint - 1], coords[end_joint - 1]))
        return np.array(connections)

    @staticmethod
    def _get_bbox_of_one_pose(coords: np.ndarray, mask: np.ndarray) -> np.ndarray:
        """Get the bounding box bordering the keypoints of a single pose"""
        coords = coords[mask, :]
        if coords.shape[0]:
            min_x, min_y, max_x, max_y = (
                coords[:, 0].min(),
                coords[:, 1].min(),
                coords[:, 0].max(),
                coords[:, 1].max(),
            )
            bbox = [min_x, min_y, max_x, max_y]
            return np.array(bbox)
        return np.zeros(0)

    def _predict_all_poses(
        self, posenet_model: tf.keras.Model, frame: np.ndarray, model_type: str
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Predict relative coordinates, confident scores and validation masks
        for all detected poses

        Args:
            posenet model: tensorflow model
            frame (np.array): image for inference
            model_type (str): specified model type (refer to modelconfig.yml)

        Returns:
            full_keypoint_coords (np.array): keypoints coordinates of detected
                poses
            full_keypoint_scores (np.array): keypoints confidence scores of
                detected poses
            full_masks (np.array): keypoints validation masks of detected poses
        """
        image, output_scale, image_size = self._create_image_from_frame(
            OUTPUT_STRIDE, frame, self.resolution, model_type
        )

        dst_scores = np.zeros((self.max_pose_detection, KEYPOINTS_NUM))
        dst_keypoints = np.zeros((self.max_pose_detection, KEYPOINTS_NUM, 2))

        pose_count = detect_keypoints(
            posenet_model,
            image,
            OUTPUT_STRIDE,
            dst_scores,
            dst_keypoints,
            model_type,
            self.score_threshold,
        )
        full_keypoint_scores = dst_scores[:pose_count]
        full_keypoint_coords = dst_keypoints[:pose_count]

        full_keypoint_rel_coords = get_keypoints_relative_coords(
            full_keypoint_coords, output_scale, image_size
        )

        full_masks = self._get_full_masks_from_keypoint_scores(full_keypoint_scores)

        return full_keypoint_rel_coords, full_keypoint_scores, full_masks

    @staticmethod
    def _create_image_from_frame(
        output_stride: int,
        frame: np.ndarray,
        input_res: Tuple[int, int],
        model_type: str,
    ) -> Tuple[tf.Tensor, np.ndarray, List[int]]:
        """Rescale raw frame and convert to tensor image for inference"""
        image_size = [frame.shape[1], frame.shape[0]]

        image, output_scale = rescale_image(
            frame,
            input_res,
            scale_factor=SCALE_FACTOR,
            output_stride=output_stride,
            model_type=model_type,
        )
        image = tf.convert_to_tensor(image)
        return image, output_scale, image_size

    @staticmethod
    def _get_full_masks_from_keypoint_scores(keypoint_scores: np.ndarray) -> np.ndarray:
        """Obtain masks for keypoints with confidence scores above the detection
        threshold
        """
        masks = keypoint_scores > MIN_PART_SCORE
        return masks



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tracker for object detector bounding boxes."""

import logging
from typing import Any, Dict, List

from peekingduck.pipeline.nodes.dabble.trackingv1.tracking_files.iou_tracker import (
    IOUTracker,
)
from peekingduck.pipeline.nodes.dabble.trackingv1.tracking_files.opencv_tracker import (
    OpenCVTracker,
)


class DetectionTracker:  # pylint: disable=too-few-public-methods
    """Tracks detection bounding boxes using the chosen algorithm.

    Args:
        config (Dict[str, Any]): Configration dict containing the following:
            tracking_type (str): Type of tracking algorithm to be used, one of
                ["iou", "mosse"].
            iou_threshold (float): Minimum IoU value to be used with the
                matching logic.
            max_lost (int): Maximum number of frames to keep "lost" tracks
                after which they will be removed. Only used in IOUTracker.

    Raises:
        ValueError: `tracking_type` is not one of ["iou", "mosse"].
        ValueError: `iou_threshold` is not within [0, 1].
        ValueError: `max_lost` is negative.
    """

    tracker_constructors = {"iou": IOUTracker, "mosse": OpenCVTracker}

    def __init__(self, config: Dict[str, Any]) -> None:
        self.logger = logging.getLogger(__name__)

        # Check threshold values
        if not 0 <= config["iou_threshold"] <= 1:
            raise ValueError("iou_threshold must be in [0, 1]")
        if config["max_lost"] < 0:
            raise ValueError("max_lost cannot be negative")

        try:
            self.tracker = self.tracker_constructors[config["tracking_type"]](config)
        except KeyError as error:
            raise ValueError("tracking_type must be one of ['iou', 'mosse']") from error

    def track_detections(self, inputs: Dict[str, Any]) -> List[int]:
        """Tracks detections using the selected algorithm.

        Args:
            inputs (Dict[str, Any]): Dictionary with keys "img", "bboxes", and
                "bbox_scores.

        Returns:
            (List[int]): Tracking IDs of the detection bounding boxes.
        """
        track_ids = self.tracker.track_detections(inputs)
        return track_ids



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Object detection class using yolo model to detect human faces
"""

import logging
from pathlib import Path
from typing import Any, Dict, List, Tuple

import cv2
import numpy as np
import tensorflow as tf
from tensorflow.python.saved_model import tag_constants


class Detector:  # pylint: disable=too-few-public-methods
    """Object detection class using yolo model to find human faces"""

    def __init__(self, config: Dict[str, Any], model_dir: Path) -> None:
        self.logger = logging.getLogger(__name__)

        self.config = config
        self.model_dir = model_dir
        self.class_labels = self._get_class_labels()
        self.yolo = self._create_yolo_model()

    def _get_class_labels(self) -> List[str]:
        classes_path = self.model_dir / self.config["weights"]["classes_file"]
        with open(classes_path, "rt", encoding="utf8") as file:
            class_labels = [c.strip() for c in file.readlines()]

        return class_labels

    def _create_yolo_model(self) -> cv2.dnn_Net:
        model_type = self.config["model_type"]
        model_path = (
            self.model_dir / self.config["weights"]["saved_model_subdir"][model_type]
        )
        model = tf.saved_model.load(str(model_path), tags=[tag_constants.SERVING])

        self.logger.info(
            "Yolo model loaded with following configs: \n\t"
            f"Model type: {self.config['model_type']}, \n\t"
            f"Input resolution: {self.config['size']}, \n\t"
            f"IDs being detected: {self.config['detect_ids']}, \n\t"
            f"Max detections per class: {self.config['max_output_size_per_class']}, \n\t"
            f"Max total detections: {self.config['max_total_size']}, \n\t"
            f"IOU threshold: {self.config['iou_threshold']}, \n\t"
            f"Score threshold: {self.config['score_threshold']}"
        )

        return model

    def predict_object_bbox_from_image(
        self, image: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Predicts face bboxes, labels and scores

        Args:
            image (np.ndarray): image in numpy array

        Returns:
            bboxes (np.ndarray): numpy array of detected bboxes
            classes (np.ndarray): numpy array of class labels
            scores (np.ndarray): numpy array of confidence scores
        """
        # 1. process inputs
        image = self._process_image(image)

        # 2. evaluate image
        bboxes, scores, classes, nums = self._evaluate_image_by_yolo(image)

        # 3. clean up return
        bboxes, scores, classes = self._shrink_dimension_and_length(
            bboxes, scores, classes, nums
        )

        # 4. convert classes into class names
        classes = np.array([self.class_labels[int(i)] for i in classes])

        return bboxes, classes, scores

    def _process_image(self, image: np.ndarray) -> np.ndarray:
        image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
        image = cv2.resize(image, (self.config["size"], self.config["size"]))
        image = np.asarray([image]).astype(np.float32) / 255.0

        return image

    def _evaluate_image_by_yolo(
        self, image: np.ndarray
    ) -> Tuple[tf.Tensor, tf.Tensor, tf.Tensor, tf.Tensor]:
        """Performs inference on a given input image.

        Args:
            image (np.ndarray): image in numpy array

        Returns:
            bboxes (tf.Tensor): EagerTensor object of detected bboxes
            scores (tf.Tensor): EagerTensor object of confidence scores
            classes (tf.Tensor): EagerTensor object of class labels
            nums (tf.Tensor): number of valid bboxes. Only nums[0] should be used.
            The rest are paddings.
        """
        infer = self.yolo.signatures["serving_default"]
        pred_bbox = infer(tf.constant(image))
        for value in pred_bbox.values():
            boxes = value[:, :, 0:4]
            pred_conf = value[:, :, 4:]

        # performs nms using model's predictions
        bboxes, scores, classes, nums = tf.image.combined_non_max_suppression(
            max_output_size_per_class=self.config["max_output_size_per_class"],
            boxes=tf.reshape(boxes, (tf.shape(boxes)[0], -1, 1, 4)),
            scores=tf.reshape(
                pred_conf, (tf.shape(pred_conf)[0], -1, tf.shape(pred_conf)[-1])
            ),
            score_threshold=self.config["score_threshold"],
            max_total_size=self.config["max_total_size"],
            iou_threshold=self.config["iou_threshold"],
        )

        return bboxes, scores, classes, nums

    def _shrink_dimension_and_length(
        self, bboxes: tf.Tensor, scores: tf.Tensor, classes: tf.Tensor, nums: tf.Tensor
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        len0 = nums[0]

        classes = classes.numpy()[0]
        classes = classes[:len0]
        # only identify objects we are interested in
        mask1 = np.isin(classes, tuple(self.config["detect_ids"]))

        scores = scores.numpy()[0]
        scores = scores[:len0]
        scores = scores[mask1]

        bboxes = bboxes.numpy()[0]
        bboxes = bboxes[:len0]
        bboxes = bboxes[mask1]

        bboxes[:, [0, 1]] = bboxes[:, [1, 0]]
        bboxes[:, [2, 3]] = bboxes[:, [3, 2]]

        return bboxes, scores, classes



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Crowd counting class using csrnet model to predict density map and crowd count
"""

import logging
import math
from pathlib import Path
from typing import Any, Dict, Tuple, Callable

import cv2
import numpy as np
import tensorflow as tf


class Predictor:  # pylint: disable=too-few-public-methods
    """Crowd counting class using csrnet model to predict density map and crowd count"""

    def __init__(self, config: Dict[str, Any], model_dir: Path) -> None:
        self.logger = logging.getLogger(__name__)

        self.config = config
        self.model_dir = model_dir

        self.csrnet = self._create_csrnet_model()

    def _create_csrnet_model(self) -> Callable:
        """
        Creates csrnet model to predict density map and crowd count
        """
        model_type = self.config["model_type"]
        model_path = self.model_dir / self.config["weights"]["model_file"][model_type]

        self.logger.info(
            "CSRNet model loaded with following configs: \n\t"
            f"Model type: {self.config['model_type']}, \n\t"
            f"Input width: {self.config['width']} \n\t"
        )

        return self._load_csrnet(model_path)

    def _load_csrnet(self, model_path: Path) -> Callable:
        self.module = tf.saved_model.load(str(model_path))
        self.saved_model = self.module.signatures["serving_default"]

        return self.saved_model

    def predict_count_from_image(self, image: np.ndarray) -> Tuple[np.ndarray, int]:
        """Predicts density map and crowd count from image.

        Args:
            image (np.ndarray): input image.

        Returns:
            density_map (np.ndarray): resized density map.
            crowd_count (int): predicted count of people.
        """
        # 1. resizes and normalizes input image
        processed_image = self._process_input(image)

        # 2. generates the predicted density map
        density_map = self.saved_model(processed_image)["y_out"].numpy()

        # 3. resizes density map and counts the number of people
        density_map, crowd_count = self._process_output(image, density_map)

        return density_map, crowd_count

    def _process_input(self, image: np.ndarray) -> tf.Tensor:
        """Resizes and normalizes an image based on the mean and standard deviation
        of Imagenet. These are the default values for models with PyTorch origins.

        Args:
            image (np.ndarray): input image.

        Returns:
            image (np.ndarray): processed image.
        """
        image = self._resize_image(image)
        image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
        image = image / 255.0
        image[:, :, 0] = (image[:, :, 0] - 0.485) / 0.229
        image[:, :, 1] = (image[:, :, 1] - 0.456) / 0.224
        image[:, :, 2] = (image[:, :, 2] - 0.406) / 0.225
        image = np.expand_dims(image, axis=0)
        image = tf.convert_to_tensor(image, dtype=tf.float32)
        return image

    def _resize_image(self, image: np.ndarray) -> np.ndarray:
        """Resizes an image based on the input width.

        Args:
            image (np.ndarray): input image.

        Returns:
            image (np.ndarray): resized image.
        """
        ratio = self.config["width"] / image.shape[1]
        dim = (self.config["width"], int(image.shape[0] * ratio))
        image = cv2.resize(image, dim, interpolation=cv2.INTER_LINEAR)
        return image

    @staticmethod
    def _process_output(image: np.ndarray, density_map: np.ndarray) -> np.ndarray:
        """Counts the number of people and resizes the output density map to match
        the original image size. The CSRNet model returns a density map that is
        1/8 the original image size. The resized density map is used to superimpose
        a heatmap over the original image.

        Args:
            density_map (np.ndarray): predicted density map.
            image (np.ndarray): input image.

        Returns:
            density_map (np.ndarray): resized density map.
            crowd_count (int): predicted count of people.
        """
        crowd_count = math.ceil(np.sum(density_map))

        density_map = density_map[0, :, :, 0]
        density_map = cv2.resize(
            density_map,
            (image.shape[1], image.shape[0]),
            interpolation=cv2.INTER_LINEAR,
        )

        return density_map, crowd_count



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Utils for CSV logging
"""

import csv
from datetime import datetime
from pathlib import Path
from typing import Any, Dict, List


class CSVLogger:
    """Node that writes data into a csv"""

    def __init__(
        self, file_path: Path, headers: List[str], logging_interval: int = 1
    ) -> None:
        self.headers = headers.copy()
        self.headers.insert(0, "Time")
        self.file_path = file_path
        self.logging_interval = logging_interval
        self.csv_file = open(self.file_path, mode="a+", newline="")
        self.writer = csv.DictWriter(self.csv_file, fieldnames=self.headers)
        self.last_write = datetime.now()

    def write(self, data_pool: Dict[str, Any], specific_data: List[str]) -> None:
        """
        Writes a row of data in a csv file

        Args:
            data_pool(dict): the data pool of the pipeline
            specific_data(list): list of data to track

        Returns:
            None
        """
        # if file is empty write header
        if self.csv_file.tell() == 0:
            self.writer.writeheader()

        content = {k: v for k, v in data_pool.items() if k in specific_data}
        curr_time = datetime.now()
        time_str = curr_time.strftime("%H:%M:%S")
        content.update({"Time": time_str})

        if (curr_time - self.last_write).seconds >= self.logging_interval:
            self.writer.writerow(content)
            self.last_write = curr_time

    def __del__(self) -> None:
        self.csv_file.close()



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Detector class to handle detection of poses for HRNet
"""

import logging
from pathlib import Path
from typing import Any, Callable, Dict, List, Tuple

import numpy as np
import tensorflow as tf

from peekingduck.pipeline.nodes.model.hrnetv1.hrnet_files.postprocessing import (
    affine_transform_xy,
    get_keypoint_conns,
    get_valid_keypoints,
    reshape_heatmaps,
    scale_transform,
)
from peekingduck.pipeline.nodes.model.hrnetv1.hrnet_files.preprocessing import (
    box2cs,
    crop_and_resize,
    project_bbox,
)
from peekingduck.utils.graph_functions import load_graph


class Detector:
    """Detector class to handle detection of poses for HRNet."""

    def __init__(self, config: Dict[str, Any], model_dir: Path) -> None:
        self.logger = logging.getLogger(__name__)

        self.config = config
        self.model_dir = model_dir
        self.resolution = config["resolution"]
        self.min_score = config["score_threshold"]

        self.hrnet = self._create_hrnet_model()

    def _inference_function(  # pylint:disable=unused-argument
        self, person_frame: np.ndarray, training: bool = False
    ) -> tf.Tensor:
        """
        When graph is frozen, we need a different way to extract the
        arrays. We use this to return the values needed. The purpose
        of this function is to make it consistent with how our regular
        HRNet is used for inference.

        The "training" argument is needed because in the usual implementation
        of HRNet has the training argument. It does nothing here.
        """
        heatmap = self.frozen_fn(tf.cast(person_frame, float))[0]
        return heatmap

    def _create_hrnet_model(self) -> Callable:
        graph_path = self.model_dir / self.config["weights"]["model_file"]
        model_nodes = self.config["MODEL_NODES"]
        self.frozen_fn = load_graph(
            str(graph_path),
            inputs=model_nodes["inputs"],
            outputs=model_nodes["outputs"],
        )
        resolution_tuple = (self.resolution["height"], self.resolution["width"])
        self.logger.info(
            "HRNet graph model loaded with following configs: \n\t"
            f"Resolution: {resolution_tuple}, \n\t"
            f"Score Threshold: {self.min_score}"
        )
        return self._inference_function

    def preprocess(
        self, frame: np.ndarray, bboxes: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray, Tuple[int, int]]:
        """Crops bboxes while preserving aspect ratio.

        Args:
            frame (np.ndarray): Input image in numpy array.
            bboxes (np.ndarray): Array of detected bboxes.

        Returns:
            (Tuple[np.ndarray, np.ndarray, Tuple[int, int]]): Array of cropped
            bboxes, transformation matrices, and original frame size.
        """
        frame = frame / 255.0
        frame_size = (frame.shape[1], frame.shape[0])
        cropped_size = (self.resolution["width"], self.resolution["height"])

        projected_bbox = project_bbox(bboxes, frame_size)
        center_bbox = box2cs(
            projected_bbox, self.resolution["width"] / self.resolution["height"]
        )
        cropped_imgs, affine_matrices = crop_and_resize(
            frame, center_bbox, cropped_size
        )

        return np.array(cropped_imgs), affine_matrices, frame_size

    def postprocess(  # pylint: disable=too-many-locals
        self,
        heatmaps: np.ndarray,
        affine_matrices: np.ndarray,
        cropped_frames_scale: List[int],
        frame_size: Tuple[int, int],
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Post processes output heatmaps to required keypoint arrays.

        Args:
            heatmaps (np.ndarray): Output heatmaps from hrnet network.
            affine_matrices (np.ndarray): transformation matrices of preprocess
                cropping.
            cropped_frames_scale (List[int]): Shape of cropped bboxes.
            frame_size (Tuple[int, int]): Size of original image.

        Returns:
            (Tuple[np.ndarray, np.ndarray, np.ndarray]): Tuple containing array
            of bboxes and pose related info, i.e., coordinates, scores, and
            connections
        """
        batch, out_h, out_w, num_joints = heatmaps.shape
        heatmaps_reshaped = reshape_heatmaps(heatmaps)

        max_idxs = np.argmax(heatmaps_reshaped, 2)
        kp_scores = np.amax(heatmaps_reshaped, 2)
        keypoints = (
            np.repeat(max_idxs, 2).reshape(batch, num_joints, -1).astype(np.float32)
        )
        keypoints[:, :, 0] = (keypoints[:, :, 0]) % out_w
        keypoints[:, :, 1] = np.floor((keypoints[:, :, 1]) / out_w)

        keypoints = scale_transform(
            keypoints, in_scale=[out_w, out_h], out_scale=cropped_frames_scale
        )
        keypoints = affine_transform_xy(keypoints, affine_matrices)
        keypoints, kp_masks = get_valid_keypoints(
            keypoints, kp_scores, batch, self.min_score
        )

        normalized_keypoints = keypoints / frame_size
        keypoint_conns = get_keypoint_conns(normalized_keypoints, kp_masks)

        return normalized_keypoints, kp_scores, keypoint_conns

    def predict(
        self, frame: np.ndarray, bboxes: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """HRnet prediction function.

        Args:
            frame (np.ndarray): Image in numpy array.
            bboxes (np.ndarray): Array of detected bboxes.

        Returns:
            Tuple[np.ndarray, np.ndarray, np.ndarray]: Tuple containing list of
            bboxes and pose related info, i.e., coordinates, scores, and
            connections
        """
        cropped_frames, affine_matrices, frame_size = self.preprocess(frame, bboxes)
        heatmaps = self.hrnet(cropped_frames, training=False)

        cropped_frames_scale = [cropped_frames.shape[2], cropped_frames.shape[1]]
        poses, kp_scores, kp_conns = self.postprocess(
            heatmaps.numpy(), affine_matrices, cropped_frames_scale, frame_size
        )

        return poses, kp_scores, kp_conns



# Modifications copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Original copyright (c) 2020 YifuZhang
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Utility functions used by FairMOT.

Modification:
- Change _get_affine_transform to always use the same x- and y- scaling
- Change _get_affine_transform to accept np.ndarray for output_size
- Use @ instead of np.dot for matrix multiplication
"""

from typing import Tuple

import cv2
import numpy as np
import torch


def gather_feat(feat: torch.Tensor, indices: torch.Tensor) -> torch.Tensor:
    """Gathers values specified by `indices` along dim=1.

    Args:
        feat (torch.Tensor): The source tensor containing the features of
            interest.
        indices (torch.Tensor): The indices of elements to gather.

    Returns:
        (torch.Tensor): The gathered features.
    """
    dim = feat.size(2)
    indices = indices.unsqueeze(2).expand(indices.size(0), indices.size(1), dim)
    feat = feat.gather(1, indices)
    return feat


def letterbox(
    image: np.ndarray,
    height: int,
    width: int,
    colour: Tuple[float, float, float] = (127.5, 127.5, 127.5),
) -> np.ndarray:
    """Resizes a rectangular image to a padded rectangular image.

    Args:
        image (np.ndarray): Image frame.
        height (int): Height of padded image.
        width (int): Width of padded image.
        colour (Tuple[float, float, float]): Colour used for padding around
            the image. (127.5, 127.5, 127.5) is chosen as it is used by the
            original project during model training.

    Returns:
        (np.ndarray): Padded rectangular image.
    """
    shape = image.shape[:2]  # shape = [height, width]
    ratio = min(float(height) / shape[0], float(width) / shape[1])
    # new_shape = [width, height]
    new_shape = (round(shape[1] * ratio), round(shape[0] * ratio))
    width_padding = (width - new_shape[0]) / 2
    height_padding = (height - new_shape[1]) / 2
    top = round(height_padding - 0.1)
    bottom = round(height_padding + 0.1)
    left = round(width_padding - 0.1)
    right = round(width_padding + 0.1)
    # resized, no border
    image = cv2.resize(image, new_shape, interpolation=cv2.INTER_AREA)
    # padded rectangular
    image = cv2.copyMakeBorder(
        image, top, bottom, left, right, cv2.BORDER_CONSTANT, value=colour
    )
    return image


def transform_coords(
    coords: np.ndarray,
    center: np.ndarray,
    scale: float,
    output_size: Tuple[float, float],
) -> np.ndarray:
    """Translates and scales the coordinate w.r.t. the original image size.

    Args:
        coords (np.ndarray): Coordinates of a bbox corner.
        center (np.ndarray): Coordinates of the center of the original image.
        scale (float): Scale between the input image fed to the model and the
            original image.
        output_size(Tuple[float, float]): Output image size by the model.

    Returns:
        (np.ndarray): Transformed coordinates.
    """
    target_coords = np.zeros_like(coords)
    matrix = _get_affine_transform(
        center, scale, 0, np.array(output_size, dtype=np.float32), inv=True
    )
    for i in range(coords.shape[0]):
        target_coords[i, :2] = _affine_transform(coords[i, :2], matrix)
    return target_coords


def transpose_and_gather_feat(
    feat: torch.Tensor, indices: torch.Tensor
) -> torch.Tensor:
    """Transposes the features and then gather the features specified by
    `indices` at dim=1.

    Args:
        feat (torch.Tensor): The tensor containing the features of interest.
        indices (torch.Tensor): The indices of elements to gather.

    Returns:
        (torch.Tensor): The transposed and gathered features.
    """
    feat = feat.permute(0, 2, 3, 1).contiguous()
    feat = feat.view(feat.size(0), -1, feat.size(3))
    feat = gather_feat(feat, indices)
    return feat


def _affine_transform(point: np.ndarray, trans_matrix: np.ndarray) -> np.ndarray:
    """Applies affine transformation to the specified coordinates.

    Args:
        point (np.ndarray): The coordinates to transform.
        trans_matrix (np.ndarray): A 2x3 affine transformation matrix.

    Returns:
        (np.ndarray): The transformed coordinates.
    """
    new_pt = np.array([point[0], point[1], 1.0], dtype=np.float32).T
    new_pt = trans_matrix @ new_pt
    return new_pt[:2]


def _get_3rd_point(point_1: np.ndarray, point_2: np.ndarray) -> np.ndarray:
    """Computes the third point required for computing affine transformation
    matrix.

    Args:
        point_1 (np.ndarray): The first point.
        point_2 (np.ndarray): The second point.

    Returns:
        (np.ndarray): The third point.
    """
    direct = point_1 - point_2
    return point_2 + np.array([-direct[1], direct[0]], dtype=np.float32)


def _get_affine_transform(  # pylint:disable=too-many-arguments
    center: np.ndarray,
    scale: float,
    rot: float,
    output_size: np.ndarray,
    shift: np.ndarray = np.array([0, 0], dtype=np.float32),
    inv: bool = False,
) -> np.ndarray:
    """Computes the affine transformation matrix described by the given
    arguments.

    Args:
        center (np.ndarray): Coordinates of the center.
        scale (float): Scale factor.
        rot (float): Rotation amount (in degrees).
        output_size (np.ndarray): Height and width of the output.
        shift (np.ndarray): Translation.
        inv (bool): Flag to determine if we should perform the inverse of the
            transformation.

    Returns:
        (np.ndarray): The affine transformation matrix.
    """
    src_w = scale
    dst_w = output_size[0]

    src_dir = _get_dir([0, src_w * -0.5], np.pi * rot / 180)
    dst_dir = np.array([0, dst_w * -0.5], np.float32)

    src = np.zeros((3, 2), dtype=np.float32)
    dst = np.zeros((3, 2), dtype=np.float32)
    src[0, :] = center + scale * shift
    src[1, :] = center + src_dir + scale * shift
    dst[0, :] = output_size * 0.5
    dst[1, :] = output_size * 0.5 + dst_dir

    src[2, :] = _get_3rd_point(src[0, :], src[1, :])
    dst[2, :] = _get_3rd_point(dst[0, :], dst[1, :])

    if inv:
        matrix = cv2.getAffineTransform(dst, src)
    else:  # pragma: no cover
        # Only inv=True is used by FairMOT
        matrix = cv2.getAffineTransform(src, dst)

    return matrix


def _get_dir(src_point: np.ndarray, rot_rad: float) -> np.ndarray:
    """Computes the direction vector.

    Args:
        src_point (np.ndarray): The coordinates of the source point.
        rot_rad (float): The amount of rotation (in radians).

    Returns:
        (np.ndarray): The direction vector.
    """
    sin = np.sin(rot_rad)
    cos = np.cos(rot_rad)

    src_result = np.array(
        [
            src_point[0] * cos - src_point[1] * sin,
            src_point[0] * sin + src_point[1] * cos,
        ]
    )

    return src_result



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
CSRNet model with model types: sparse and dense
"""

import logging
from typing import Dict, Any, Tuple

import numpy as np

from peekingduck.pipeline.nodes.model.csrnetv1.csrnet_files.predictor import Predictor
from peekingduck.weights_utils import checker, downloader, finder


class CsrnetModel:  # pylint: disable=too-few-public-methods
    """CSRNet model with model types: sparse and dense"""

    def __init__(self, config: Dict[str, Any]) -> None:
        super().__init__()
        self.logger = logging.getLogger(__name__)

        weights_dir, model_dir = finder.find_paths(
            config["root"], config["weights"], config["weights_parent_dir"]
        )

        # check for csrnet weights, if none then download into weights folder
        if not checker.has_weights(weights_dir, model_dir):
            self.logger.info("---no weights detected. proceeding to download...---")
            downloader.download_weights(weights_dir, config["weights"]["blob_file"])
            self.logger.info(f"---weights downloaded to {weights_dir}.---")

        self.predictor = Predictor(config, model_dir)

    def predict(self, frame: np.ndarray) -> Tuple[np.ndarray, int]:
        """Predicts density map and crowd count from frame.

        Args:
            frame (np.ndarray): input frame.

        Returns:
            density_map (np.ndarray): predicted density map.
            crowd_count (int): predicted count of people.
        """
        assert isinstance(frame, np.ndarray)

        # return density_map, crowd_count
        return self.predictor.predict_count_from_image(frame)



# Modifications copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Original copyright (c) 2017 TU Berlin, Communication Systems Group
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so.
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Tracking-by-detection using IoU Tracker."""

from collections import OrderedDict
from typing import Any, Dict, List, Tuple

import numpy as np

from peekingduck.pipeline.nodes.dabble.trackingv1.tracking_files.track import Track
from peekingduck.pipeline.nodes.dabble.trackingv1.tracking_files.utils import (
    iou_tlwh,
    xyxyn2tlwh,
)


class IOUTracker:
    """Simple tracking class based on Intersection over Union (IoU) of bounding
    boxes.

    This method is based on the assumption that the detector produces a
    detection per frame for every object to be tracked. Furthermore, it is
    assumed that detections of an object in consecutive frames have an
    unmistakably high overlap IoU which is commonly the case when using
    sufficiently high frame rates.

    References:
        High-Speed Tracking-by-Detection Without Using Image Information:
        http://elvera.nue.tu-berlin.de/files/1517Bochinski2017.pdf

        Inference code adapted from
        https://github.com/adipandas/multi-object-tracker
    """

    def __init__(self, config: Dict[str, Any]) -> None:
        self.iou_threshold = config["iou_threshold"]
        self.max_lost = config["max_lost"]
        self.next_track_id = 0

        self.tracks: "OrderedDict[int, Track]" = OrderedDict()

    def track_detections(self, inputs: Dict[str, Any]) -> List[int]:
        """Initializes and updates tracker on each frame.

        Args:
            inputs (Dict[str, Any]): Dictionary with keys "img" and "bboxes".

        Returns:
            (List[int]]): List of track IDs.
        """
        frame = inputs["img"]
        frame_size = frame.shape[:2]
        tlwhs = xyxyn2tlwh(inputs["bboxes"], *frame_size)

        tracks = self.update(list(tlwhs))
        track_ids = self._order_track_ids_by_bbox(tlwhs, tracks)

        return track_ids

    def update(self, detections: np.ndarray) -> List[Track]:
        """Updates the tracker. Creates new tracks for untracked objects,
        updates tracked objects with the new class ID and bounding box
        coordinates. Removes tracks which have not been detected for longer
        than the `max_lost` threshold.

        Args:
            detections (np.ndarray): Bounding box coordinates for each of
                detection. The bounding box has the format (t, l, w, h) where
                (t, l) is the top-left corner, w is the width, and h is the
                height.

        Returns:
            (List[Track]): All tracked detections in the current frame.
        """
        track_ids = list(self.tracks.keys())

        updated_tracks = []
        for track_id in track_ids:
            if detections:
                idx, best_match, best_iou = self.get_best_match_by_iou(
                    detections, self.tracks[track_id].bbox
                )
                if best_iou >= self.iou_threshold:
                    self._update_track(track_id, best_match, best_iou)
                    updated_tracks.append(track_id)
                    del detections[idx]
            if not updated_tracks or track_id != updated_tracks[-1]:
                self.tracks[track_id].lost += 1
                if self.tracks[track_id].lost > self.max_lost:
                    self._remove_track(track_id)
        for tlwh in detections:
            self._add_track(tlwh)
        outputs = self._get_tracks()
        return outputs

    def _add_track(self, bbox: np.ndarray) -> None:
        """Adds a newly detected object to the list of tracked detections.

        Args:
            bbox (np.ndarray): Bounding box with format (t, l, w, h) where
                (t, l) is the top-left corner, w is the width, and h is the
                height.
        """
        self.tracks[self.next_track_id] = Track(self.next_track_id, bbox)
        self.next_track_id += 1

    def _get_tracks(self) -> List[Track]:
        """All tracked detections in the current frame."""
        return [track for _, track in self.tracks.items() if track.lost == 0]

    def _remove_track(self, track_id: int) -> None:
        """Removes the specified track. Typically called when the track has not
        been detected in the frame for longer than `max_lost` consecutive
        frames.

        Args:
            track_id (int): The track to delete.
        """
        del self.tracks[track_id]

    def _update_track(self, track_id: int, bbox: np.ndarray, iou_score: float) -> None:
        """Updates the specified tracked detection.

        Args:
            track_id (int): ID of the tracked detection.
            bbox (np.ndarray): Bounding box coordinates with (t, l, w, h)
                format where (t, l) is the top-left corner, w is the width, and
                h is the height.
            iou_score (float): Intersection-over-Union between the current
                detection bounding box and its last detected bounding box.
        """
        self.tracks[track_id].update(bbox, iou_score)

    @staticmethod
    def get_best_match_by_iou(
        detections: List[np.ndarray], tracked_bbox: np.ndarray
    ) -> Tuple[int, np.ndarray, float]:
        """Finds the best match between all the current detections and the
        specified tracked bounding box. Best match is the pair with the largest
        IoU value.

        Args:
            detections (List[np.ndarray]): List of tuples containing the
                bounding box coordinates for each of the detection. The
                bounding box has the format (t, l, w, h) where (t, l) is the
                top-left corner, w is the width, and h is the height.
            tracked_bbox (np.ndarray): The specified tracked bounding box.

        Returns:
            (Tuple[int, np.ndarray, float]): The index, bounding box of the
            best match current detection, and the IoU value.
        """
        detection_and_iou = lambda det: (det, iou_tlwh(tracked_bbox, det))
        idx, (best_match, best_iou) = max(
            enumerate(map(detection_and_iou, detections)),
            key=lambda x: x[1][1],
        )
        return idx, best_match, best_iou

    @staticmethod
    def _order_track_ids_by_bbox(bboxes: np.ndarray, tracks: List[Track]) -> List[int]:
        """Extracts the track IDs and orders them by their respective bounding
        boxes.

        Args:
            bboxes (np.ndarray): Detection bounding boxes with (t, l, w, h)
                format where (t, l) is the top-left corner, w is the width, and
                h is the height.
            tracks (List[Track]): List of tracked detections.

        Returns:
            (List[int]): Track IDs of the detections in the current frame.
        """
        bbox_to_track_id = {tuple(track.bbox): track.track_id for track in tracks}
        track_ids = [bbox_to_track_id[bbox] for bbox in map(tuple, bboxes)]
        return track_ids



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""🔲 Multi-task Cascaded Convolutional Networks for face detection. Works best
with unmasked faces.
"""

from typing import Any, Dict

import numpy as np

from peekingduck.pipeline.nodes.model.mtcnnv1 import mtcnn_model
from peekingduck.pipeline.nodes.node import AbstractNode


class Node(AbstractNode):
    """Initializes and uses the MTCNN model to infer bboxes from an image
    frame.

    The MTCNN node is a single-class model capable of detecting human faces. To
    a certain extent, it is also capable of detecting bounding boxes around
    faces with face masks (e.g. surgical masks).

    Inputs:
        |img_data|

    Outputs:
        |bboxes_data|

        |bbox_scores_data|

        |bbox_labels_data|

    Configs:
        weights_parent_dir (:obj:`Optional[str]`): **default = null**. |br|
            Change the parent directory where weights will be stored by
            replacing ``null`` with an absolute path to the desired directory.
        min_size (:obj:`int`): **default = 40**. |br|
            Minimum height and width of face in pixels to be detected.
        scale_factor (:obj:`float`): **[0, 1], default = 0.709**. |br|
            Scale factor to create the image pyramid. A larger scale factor
            produces more accurate detections at the expense of inference
            speed.
        network_thresholds (:obj:`List`):
            **[0, 1], default = [0.6, 0.7, 0.7]**. |br|
            Threshold values for the Proposal Network (P-Net), Refine Network
            (R-Net) and Output Network (O-Net) in the MTCNN model.

            Calibration is performed at each stage in which bounding boxes with
            confidence scores less than the specified threshold are discarded.
        score_threshold (:obj:`float`): **[0, 1], default = 0.7**. |br|
            Bounding boxes with confidence scores less than the specified
            threshold in the final output are discarded.

    References:
        Joint Face Detection and Alignment using Multi-task Cascaded
        Convolutional Networks:
        https://arxiv.org/ftp/arxiv/papers/1604/1604.02878.pdf

        Model weights trained by https://github.com/blaueck/tf-mtcnn

    .. versionchanged:: 1.2.0
        ``mtcnn_min_size`` is renamed to ``min_size``.

    .. versionchanged:: 1.2.0
        ``mtcnn_factor`` is renamed to ``scale_factor``.

    .. versionchanged:: 1.2.0
        ``mtcnn_thresholds`` is renamed to ``network_thresholds``.

    .. versionchanged:: 1.2.0
        ``mtcnn_score`` is renamed to ``score_threshold``.
    """

    def __init__(self, config: Dict[str, Any] = None, **kwargs: Any) -> None:
        super().__init__(config, node_path=__name__, **kwargs)
        self.model = mtcnn_model.MtcnnModel(self.config)

    def run(self, inputs: Dict[str, Any]) -> Dict[str, Any]:
        """Reads the image input and returns the bboxes, scores and labels of
        faces detected.

        Args:
            inputs (dict): Dictionary of inputs with key "img".

        Returns:
            outputs (dict): Outputs in dictionary format with keys "bboxes",
            "bbox_scores", and "bbox_labels".
        """
        bboxes, scores, _, classes = self.model.predict(inputs["img"])
        bboxes = np.clip(bboxes, 0, 1)

        outputs = {"bboxes": bboxes, "bbox_scores": scores, "bbox_labels": classes}
        return outputs



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Checks for model weights in weights folder.
"""

from pathlib import Path


def has_weights(weights_dir: Path, model_dir: Path) -> bool:
    """Checks for model weight paths from weights folder.

    Args:
        weights_dir (:obj:`Path`): Path to where all weights are stored.
        model_dir (:obj:`Path`): Path to where weights for a model are stored.

    Returns:
        (:obj:`bool`): ``True`` if specified files/directories in
        ``weights_dir`` exist, else ``False``.
    """
    if not weights_dir.exists():
        weights_dir.mkdir()
        return False
    return model_dir.exists()
